# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

import bpy
from bpy.props import *
from bpy.types import PropertyGroup

def update_fx(self, context):
    scn = context.scene
    scn.animax_update_tag = True
    return None

def only_curve(self, object):
    return object.type == 'CURVE'

def only_geo_node(self, object):
    return object.type == 'GEOMETRY' and 'ANIMAX_' not in object.name

def coll_poll(self, object):
    return object.name != 'ANIMAX_Containers'
############################################################################################
##################################### Properties ###########################################
############################################################################################

class AnimaxExposedFX(PropertyGroup):
    fx : StringProperty()
    id : StringProperty()
    original_name : StringProperty()
    ui_name : StringProperty(
        name = 'UI Name',
        description = 'Name of the parameter in the UI'
    )
    description : StringProperty(
        name = 'Description',
        description = 'Description of the parameter'
    )

class AnimaxFX(PropertyGroup):
    name : StringProperty(
        name = 'Name',
        description = 'Name of the Effect'
    )
    ui_name : StringProperty()
    type : StringProperty()
    enabled : BoolProperty(
        name = 'Enable FX',
        default = True,
        update = update_fx,
        description = 'Enable/Disable this effect'
    )
    offset : EnumProperty(
        name = 'Offset Method',
        items = (
            (
                'POSITION',
                'Position',
                'Offset based on the original position of the elements',
                'OUTLINER_OB_POINTCLOUD',
                0
            ),
            (
                'EFFECTOR',
                'Effector',
                'Offset based on the proximity to an object',
                'SNAP_ON',
                1
            ),
            (
                'TEXTURE',
                'Texture',
                'Offset based on a grayscale texture',
                'NODE_TEXTURE',
                2
            ),
        ),
        default = 'POSITION',
        update = update_fx,
        description = 'The method used to offset the elements'
    )
    sync : BoolProperty(
        name = 'Synchronize',
        default = False,
        update = update_fx,
        description = 'Synchronize with a previous effect'
    )
    sync_with : StringProperty(
        name = 'Sync With',
        update = update_fx,
        description = 'Synchronize with this effect'
    )
    easing : EnumProperty(
        name = 'Easing Functions',
        items = (
            ('LINEAR','Linear','', 'IPO_LINEAR', 0),
            ('IN','Ease In','', 'IPO_EASE_IN', 1),
            ('OUT','Ease Out','', 'IPO_EASE_OUT', 2),
            ('IN_OUT','Ease In Out','', 'IPO_EASE_IN_OUT', 3),
            ('OUT_IN','Ease Out In','', 'IPO_EASE_IN_OUT', 4),
            ('BACK_IN','Back In','', 'IPO_BACK', 5),
            ('BACK_OUT','Back Out','', 'IPO_BACK', 6),
            ('BACK_IN_OUT','Back In Out','', 'IPO_BACK', 7),
            ('ELASTIC_IN','Elastic In','', 'IPO_ELASTIC', 8),
            ('ELASTIC_OUT','Elastic Out','', 'IPO_ELASTIC', 9),
            ('ELASTIC_IN_OUT','Elastic In Out','', 'IPO_ELASTIC', 10),
            ('BOUNCE_IN','Bounce In','', 'IPO_BOUNCE', 11),
            ('BOUNCE_OUT','Bounce Out','', 'IPO_BOUNCE', 12),
            ('BOUNCE_IN_OUT','Bounce In Out','', 'IPO_BOUNCE', 13),
            ('INVERT','Invert','', 'ARROW_LEFTRIGHT', 14),
        ),
        default = 'LINEAR',
        update = update_fx,
        description = 'The math function used for easing (Tweening)'
    )
    easing_smooth_step : FloatProperty(
        name = 'Smooth Step',
        default = 2.0,
        min = 1.0,
        max = 10.0,
        description = 'Smooth the curve'
    )
    easing_frequency : IntProperty(
        name = 'Frequency',
        default = 1,
        min = 1,
        max = 10000,
        description = 'Frequency'
    )
    easing_amplitude : FloatProperty(
        name = 'Amplitude',
        default = 1.5,
        min = 1.0,
        max = 10000.0,
        description = 'Amplitude'
    )
    start_frame : IntProperty(
        name = 'Start Frame',
        default = 1,
        min = -1000000,
        max = 1000000,
        description = 'First frame of the animation'
    )
    duration : IntProperty(
        name = 'Duration',
        default = 48,
        min = 1,
        max = 1000000,
        description = 'Duration of the effect in frames'
    )
    time : FloatProperty(
        name = 'Time',
        default = 0.0,
        min = 0.0,
        max = 1.0,
        description = 'Time'
    )
    control_time : BoolProperty(
        name = 'Control the Time',
        default = False,
        description = 'Control the time manually (Time Remapping)'
    )
    invert_time : BoolProperty(
        name = 'Invert the time',
        default = False,
        description = 'Invert the time (works only with the first effect)'
    )
    repeat_time : IntProperty(
        name = 'Repeat the Time',
        default = 1,
        min = 1,
        max = 1000000,
        description = 'Repeat the Time (Loop)'
    )
    ping_pong : BoolProperty(
        name = 'Ping-Pong',
        default = False,
        description = 'Repeat the Time in a back and forth pattern'
    )
    random_offset : FloatProperty(
        name = 'Randomize',
        default = 0.0,
        min = 0.0,
        max = 1.0,
        description = 'Randomize the offset'
    )
    random_offset_seed : IntProperty(
        name = 'Random Seed',
        default = 0,
        min = 0,
        max = 1000000,
        description = 'Random Seed'
    )
    offset_falloff : FloatProperty(
        name = 'Falloff',
        default = 1.0,
        min = 1.0,
        max = 1000000.0,
        description = 'Falloff'
    )
    offset_direction : EnumProperty(
        name = 'Direction',
        items = (
            ('X','X',''),
            ('Y','Y',''),
            ('Z','Z',''),
            ('Center','Center',''),
        ),
        default = 'X',
        update = update_fx,
        description = 'Direction of the offset'
    )
    invert_direction : BoolProperty(
        name = 'Invert the direction',
        default = False,
        description = 'Invert the direction of the offset'
    )
    easing_type : EnumProperty(
        name = 'Easing Type',
        items = (
            ('FUNCTION','Function','Math function', 'CON_TRANSFORM', 0),
            ('CURVE','Curve','Custom curve', 'IPO_BEZIER', 1),
            ('PING_PONG','Ping-Pong','Back and forth', 'INVERSESQUARECURVE', 2),
        ),
        default = 'FUNCTION',
        update = update_fx,
        description = 'The method used for interpolation'
    )
    easing_curve_factor : FloatProperty(
        name = 'Factor',
        default = 1.0,
        min = 0.0,
        max = 1.0,
        description = 'Factor'
    )
    effector_type : EnumProperty(
        name = 'Effector Type',
        items = (
            ('Origine','Origine',''),
            ('Direction','Direction',''),
            ('Points','Points',''),
            ('Edges','Edges',''),
            ('Faces','Faces',''),
        ),
        default = 'Origine',
        update = update_fx,
        description = 'The method used for proximity'
    )
    effector_falloff : FloatProperty(
        name = 'Falloff',
        default = 1.0,
        min = 1.0,
        max = 1000000.0,
        description = 'Falloff'
    )
    effector_distance : FloatProperty(
        name = 'Distance',
        default = 1.0,
        min = 0.00001,
        max = 1000000.0,
        description = 'Minimum distance bettween the effector and the element'
    )
    effector_object : PointerProperty(
        name = 'Object',
        type = bpy.types.Object,
        update = update_fx,
        description = 'The Object to be used as an effector'
    )
    use_effector_scale : BoolProperty(
        name = "Use Effector's Scale",
        default = True,
        description = 'Use the scale of the efctor as the minimum distance'
    )
    location : FloatVectorProperty(
        name = 'Location',
        subtype = 'XYZ',
        description = 'Location'
    )
    rotation : FloatVectorProperty(
        name = 'Rotation',
        subtype = 'EULER',
        description = 'Rotation'
    )
    scale : FloatVectorProperty(
        name = 'Scale',
        default = (0.0, 0.0, 0.0),
        subtype = 'XYZ',
        description = 'Scale'
    )
    random_location : FloatVectorProperty(
        name = 'Location',
        subtype = 'XYZ',
        description = 'Randomize the Location'
    )
    random_rotation : FloatVectorProperty(
        name = 'Rotation',
        subtype = 'EULER',
        description = 'Randomize the Rotation'
    )
    random_scale : FloatVectorProperty(
        name = 'Scale',
        default = (0.0, 0.0, 0.0),
        subtype = 'XYZ',
        min = 0.0,
        max = 1000000.0,
        description = 'Randomize the Scale'
    )
    random_location_seed : IntProperty(
        name = 'Random Seed',
        default = 0,
        min = 0,
        max = 1000000,
        description = 'Random Seed'
    )
    random_rotation_seed : IntProperty(
        name = 'Random Seed',
        default = 0,
        min = 0,
        max = 1000000,
        description = 'Random Seed'
    )
    random_scale_seed : IntProperty(
        name = 'Random Seed',
        default = 0,
        min = 0,
        max = 1000000,
        description = 'Random Seed'
    )
    explode_location : BoolProperty(
        name = 'Explode Location',
        default = False,
        description = 'Push the objects away from the center'
    )
    replace_location : BoolProperty(
        name = 'Replace Location',
        default = False,
        description = 'Replace the location from the previous effect(s) by the current location'
    )
    replace_rotation : BoolProperty(
        name = 'Replace Rotation',
        default = False,
        description = 'Replace the rotation from the previous effect(s) by the current rotation'
    )
    replace_scale : BoolProperty(
        name = 'Replace Scale',
        default = False,
        description = 'Replace the scale from the previous effect(s) by the current scale'
    )
    uniform_scale : BoolProperty(
        name = 'Uniform Scale',
        default = True,
        description = 'The axis (x, y, z) will have the same value'
    )
    uniform_scale_value : FloatProperty(
        name = 'Scale',
        default = 0.0,
        min = -100000.0,
        max = 1000000.0,
        description = 'Scale'
    )
    uniform_random_scale : FloatProperty(
        name = 'Scale',
        default = 0.0,
        min = -100000.0,
        max = 1000000.0,
        description = 'Scale'
    )
    scale_fade_out : BoolProperty(
        name = 'Fade Out',
        default = False,
        description = 'Fade Out'
    )
    fade_out_smooth_step : FloatProperty(
        name = 'Smooth Step',
        default = 0.5,
        min = 0.0,
        max = 1.0,
        description = 'Smooth Step'
    )
    noise_frequency : FloatProperty(
        name = 'Frequency',
        default = 100.0,
        min = 0.0,
        max = 1000000.0,
        description = 'Frequency'
    )
    noise_scale : FloatProperty(
        name = 'Scale',
        default = 1.0,
        min = 0.0,
        max = 1000000.0,
        description = 'Scale'
    )
    noise_detail : FloatProperty(
        name = 'Detail',
        default = 0.0,
        min = 0.0,
        max = 15.0,
        description = 'Detail'
    )
    noise_roughness : FloatProperty(
        name = 'Roughness',
        default = 0.0,
        min = 0.0,
        max = 1.0,
        description = 'Roughness'
    )
    noise_distortion : FloatProperty(
        name = 'Distortion',
        default = 0.0,
        min = 0.0,
        max = 10000.0,
        description = 'Distortion'
    )
    random_seed : IntProperty(
        name = 'Random Seed',
        default = 0,
        min = 0,
        max = 1000000,
        description = 'Random Seed'
    )
    continuous_animation : BoolProperty(
        name = 'Continuous Animation',
        default = True,
        description = 'The animation is not bound by the duration'
    )
    wave_phase : FloatProperty(
        name = 'Phase Offset',
        default = 0.0,
        min = -1000000.0,
        max = 1000000.0,
        description = 'Phase Offset'
    )
    follow_world_position : BoolProperty(
        name = 'Use World Position',
        default = False,
        description = 'Use the world position of the curve'
    )
    follow_reverse_curve : BoolProperty(
        name = 'Reverse Curve',
        default = False,
        description = 'Reverse the direction of the curve'
    )
    align_rotation : BoolProperty(
        name = 'Align To Curve',
        default = False,
        update = update_fx,
        description = 'Align the rotation of the elements to their trajectory'
    )
    invert_axis : BoolProperty(
        name = 'Invert Axis',
        default = False,
        description = 'Invert Axis'
    )
    follow_resolution : IntProperty(
        name = 'Curve Resolution',
        default = 256,
        min = 0,
        max = 1000000,
        description = 'Resolution of the curve'
    )
    follow_object : PointerProperty(
        name = 'Object',
        type = bpy.types.Object,
        update = update_fx,
        poll = only_curve,
        description = 'The curve that the elements will follow'
    )
    num_turns : FloatProperty(
        name = 'Number of turns',
        default = 0.0,
        min = -1000000.0,
        max = 1000000.0,
        description = 'Number of rotations around the curve'
    )
    axis_enum: EnumProperty(
        name = 'Axis',
        items = (
            ('0', 'X', ''),
            ('1', 'Y', ''),
            ('2', 'Z', '')
        ),
        description = 'Axis to align to the vector'
    )
    attractor : PointerProperty(
        name = 'Object',
        type = bpy.types.Object,
        update = update_fx,
        description = 'The object that the elements will be attracted to'
    )
    velocity : FloatVectorProperty(
        name = 'Velocity',
        subtype = 'XYZ',
        description = 'Velocity'
    )
    gravity : FloatVectorProperty(
        name = 'Gravity',
        subtype = 'XYZ',
        default = (0.0, 0.0, 10.0),
        description = 'Gravity'
    )
    time_multiplier : FloatProperty(
        name = 'Time Multiplier',
        default = 1.0,
        min = 0.00001,
        max = 1000000.0,
        description = 'Multiply the time by this factor'
    )
    repulsor : PointerProperty(
        name = 'Object',
        type = bpy.types.Object,
        update = update_fx,
        description = 'The object that the elements will avoid'
    )
    force : FloatVectorProperty(
        name = 'Force',
        subtype = 'XYZ',
        description = 'Repulsion Strength'
    )
    distance_vec : FloatVectorProperty(
        name = 'Distance',
        subtype = 'XYZ',
        description = 'Distance'
    )
    distance : FloatProperty(
        name = 'Distance',
        default = 5.0,
        min = 0.0,
        max = 1000000.0,
        description = 'Minimum distance'
    )
    repulse_method: EnumProperty(
        name = 'Proximity Method',
        items = (
            ('0', 'Origine', ''),
            ('1', 'Points', ''),
            ('2', 'Edges', ''),
            ('3', 'Faces', '')
        ),
        description = 'The method used for proximity'
    )
    col_ob_switch: EnumProperty(
        name = 'Target Type',
        items = (
            ('0', 'Object', ''),
            ('1', 'Collection', '')
        ),
        description = 'Type of the target, it could be a single mesh object or a collection'
    )
    target_ob : PointerProperty(
        name = 'Target',
        type = bpy.types.Object,
        update = update_fx,
        description = 'Target object'
    )
    target_col : PointerProperty(
        name = 'Target Collection',
        type = bpy.types.Collection,
        update = update_fx,
        description = 'The collection of objects to be used as a target'
    )
    ray_direction : FloatVectorProperty(
        name = 'Ray Direction',
        subtype = 'XYZ',
        default = (0.0, 0.0, -1.0),
        description = 'Direction of the ray(s)'
    )
    ray_length : FloatProperty(
        name = 'Ray Length',
        default = 200.0,
        min = 0.0,
        max = 1000000.0,
        description = 'Length of the ray(s)'
    )
    factor : FloatProperty(
        name = 'Factor',
        default = 1.0,
        min = 0.0,
        max = 1.0,
        description = 'Factor'
    )
    fade_out_missed : BoolProperty(
        name = 'Fade out unprojected',
        default = False,
        description = 'Fade out the elements that are not projected'
    )
    collapse_axis : FloatVectorProperty(
        name = 'Collapse Axis',
        subtype = 'XYZ',
        default = (0.0, 0.0, 0.0),
        min = 0.0,
        max = 1.0,
        description = 'Collapse the elements'
    )
    collapse_offset : FloatVectorProperty(
        name = 'Offset',
        subtype = 'XYZ',
        default = (0.0, 0.0, 0.0),
        description = 'Offset the center of the elements'
    )
    collapse_rotation : FloatVectorProperty(
        name = 'Rotation',
        subtype = 'EULER',
        description = 'Rotate the elements around the median point '
    )
    center_ob : PointerProperty(
        name = 'Center Object',
        type = bpy.types.Object,
        update = update_fx,
        description = 'The origin of the object will be used as the center of rotation'
    )
    coord_rotation : FloatVectorProperty(
        name = 'Coordinates Rotation',
        subtype = 'EULER',
        description = 'Rotation'
    )
    details_scale : FloatProperty(
        name = 'Details Scale',
        default = 0.0,
        min = 0.0,
        max = 10000.0,
        description = 'Details Scale'
    )
    
    texture_type : EnumProperty(
        name = 'Texture Type',
        items = (
            ('LINEAR','Gradient',''),
            ('RADIAL','Radial',''),
            ('NOISE','Noise',''),
            ('IMAGE','Image',''),
        ),
        default = 'LINEAR',
        update = update_fx,
        description = 'The texture that will drive the offset'
    )
    image_texture : PointerProperty(
        name = 'Image',
        type = bpy.types.Image,
        update = update_fx,
        description = 'The luminance of the image will guide the offset'
    )
    texture_rotation : FloatVectorProperty(
        name = 'Rotation',
        subtype = 'EULER',
        description = 'Rotation'
    )
    texture_location : FloatVectorProperty(
        name = 'Location',
        subtype = 'XYZ',
        description = 'Location'
    )
    texture_scale : FloatVectorProperty(
        name = 'Scale',
        subtype = 'XYZ',
        default = (1.01, 1.01, 1.01),
        description = 'Scale'
    )
    noise_texture_scale : FloatProperty(
        name = 'Scale',
        default = 5.0,
        min = 0.0,
        max = 10000.0,
        description = 'Scale'
    )
    noise_texture_detail : FloatProperty(
        name = 'Detail',
        default = 0.0,
        min = 0.0,
        max = 15.0,
        description = 'Detail'
    )
    noise_texture_roughness : FloatProperty(
        name = 'Roughness',
        default = 0.0,
        min = 0.0,
        max = 1.0,
        description = 'Roughness'
    )
    noise_texture_distortion : FloatProperty(
        name = 'Distortion',
        default = 0.0,
        min = 0.0,
        max = 10000.0,
        description = 'Distortion'
    )
    noise_texture_frequency : FloatProperty(
        name = 'Frequency',
        default = 0.0,
        min = 0.0,
        max = 1000000.0,
        description = 'Frequency'
    )

class AnimaxItem(PropertyGroup):
    name : StringProperty(
        name = 'Name',
        description = 'Name of the item'
    )
    ui_name : StringProperty()
    type : StringProperty()
    geo_node_container : PointerProperty(type = bpy.types.Object)
    fx_enabled : BoolProperty(
        name = 'Enable FX',
        default = True,
        description = 'Enable/Disable the effects for this item'
    )
    control_time : BoolProperty(
        name = 'Control the time',
        default = False,
        description = 'Control the time manually for all the effects (Time Remapping)',
        update = update_fx,
    )
    frame : FloatProperty(
        name = 'Frame',
        default = 1.0,
        min = -1000000.0,
        max = 1000000.0,
        description = 'Use this value as the Current Frame'
    )
    realize_instances : BoolProperty(
        name = 'Realize Instances',
        default = False,
        description = 'Convert the Instances to real geometry.\n'
        'Useful for passing attributes from Geometry Nodes to the Shader Nodes',
        update = update_fx,
    )
    merge_by_distance : BoolProperty(
        name = 'Merge by distance',
        default = True,
        description = 'Merge the vertices that are very close (remove doubles).\n'
        'For a cleaner geometry',
        update = update_fx,
    )
    collection : PointerProperty(
        name = 'Collection',
        type = bpy.types.Collection,
        update = update_fx,
        description = 'Select a Collection',
        poll = coll_poll
    )
    custom_ng : PointerProperty(
        name = 'Node Group',
        type = bpy.types.NodeTree,
        poll = only_geo_node,
        description = 'Select a Node-group',
        update = update_fx,
    )
    use_local_location : BoolProperty(
        name = 'Use Local Location',
        default = False,
        description = 'Move the elements in Local Space'
    )
    use_local_rotation : BoolProperty(
        name = 'Use Local Rotation',
        default = True,
        description = 'Rotate the elements in Local Space'
    )
    use_local_scale : BoolProperty(
        name = 'Use Local Scale',
        default = True,
        description = 'Scale the elements in Local Space'
    )
    fx : CollectionProperty(type=AnimaxFX)
    fx_index : IntProperty(
        name = 'Effect',
        description = 'Name of the Effect'
    )
    exposed_fx : CollectionProperty(type=AnimaxExposedFX)
    exposed_fx_index : IntProperty()
    delete_invisible : BoolProperty(
        name = 'Delete Invisible',
        default = True,
        description = 'Delete the instances that their scale is zero (Faded out).\n'
        'To improve the performance'
    )
   
class AnimaxGroup(PropertyGroup):
    coll : CollectionProperty(type=AnimaxItem)
    index : IntProperty(
        name = 'Item',
        description = 'Name of the Item'
    )
    containers_coll : PointerProperty(
        name = 'Containers Collection',
        type = bpy.types.Collection,
        description = 'Containers Collection'
    )


############################################################################################
################################ Register/Unregister #######################################
############################################################################################
    
classes = (    
    AnimaxFX,
    AnimaxExposedFX,
    AnimaxItem,
    AnimaxGroup
)

def register():
    from bpy.utils import register_class
    for cls in classes:
        register_class(cls)
        
    bpy.types.Scene.animax_items = PointerProperty(type = AnimaxGroup)
    bpy.types.Scene.animax_fx = PointerProperty(type = AnimaxFX)
    bpy.types.Scene.animax_exposed_fx = PointerProperty(type = AnimaxExposedFX)
    

def unregister():
    from bpy.utils import unregister_class
    for cls in reversed(classes):
        unregister_class(cls)
        
    del bpy.types.Scene.animax_items
    del bpy.types.Scene.animax_fx
    del bpy.types.Scene.animax_exposed_fx

register()