# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

import bpy

easing_functions = {
    'LINEAR': '',
    'IN': 'ANIMAX_Ease_In_NG',
    'OUT': 'ANIMAX_Ease_Out_NG',
    'IN_OUT': 'ANIMAX_Ease_In_Out_NG',
    'OUT_IN': 'ANIMAX_Ease_Out_In_NG',
    'BACK_IN': 'ANIMAX_Back_In_NG',
    'BACK_OUT': 'ANIMAX_Back_Out_NG',
    'BACK_IN_OUT': 'ANIMAX_Back_In_Out_NG',
    'ELASTIC_IN': 'ANIMAX_Elastic_In_NG',
    'ELASTIC_OUT': 'ANIMAX_Elastic_Out_NG',
    'ELASTIC_IN_OUT': 'ANIMAX_Elastic_In_Out_NG',
    'BOUNCE_IN': 'ANIMAX_Bounce_In_NG',
    'BOUNCE_OUT': 'ANIMAX_Bounce_Out_NG',
    'BOUNCE_IN_OUT': 'ANIMAX_Bounce_In_Out_NG',
    'INVERT': 'ANIMAX_Invert_Time_NG',
}

# Margine between the nodes
x_margine, y_margine = [400.0, 500.0]

# Load a GeometryNodes node group (append from a blend file)
def load_ng(ng_dir, ng_name):
    with bpy.data.libraries.load(ng_dir, link=False) as (data_src, data_dst):
        data_dst.node_groups = [ng_name]
    return bpy.data.node_groups.get(ng_name)

# Find the effect node (if present)
def get_fx_node(nodes, ng_name):
    fx_node = nodes.get('Effect')
    if fx_node is None:
        return None
    if not fx_node.type == 'GROUP':
        nodes.remove(fx_node)
        return None
    if not fx_node.node_tree.name == ng_name:
        nodes.remove(fx_node)
        return None
    return fx_node

# Add FX Node
def add_fx_node(nodes, links, fx, ng_name, ng_dir):
    node_groups = bpy.data.node_groups
    fx_node = nodes.new(type = 'GeometryNodeGroup')
    fx_node.name = 'Effect'
    fx_node.label = fx.type
    ng = node_groups.get(ng_name)
    if ng is None:
        ng = load_ng(ng_dir, ng_name)
    if ng is None:
        return
        
    fx_node.node_tree = ng
    input = nodes['Group Input']
    output = nodes['Output']
    
    props = [
        'Center Vectors',
        'Frame',
        'Instances',
        'Position',
        'Previous Location',
        'Previous Rotation',
        'Previous Scale'
    ]
    
    for i in props:
        if not i in fx_node.inputs.keys():
            continue
        links.new(input.outputs[i], fx_node.inputs[i])
    
    for i in ['Location', 'Rotation', 'Scale']:
        links.new(fx_node.outputs[i], output.inputs[i])
    return fx_node

# Add Offset Node
def add_offset_node(scn, item, nodes, fx, ng_name, ng_dir):
    node_groups = bpy.data.node_groups
    offset = nodes.get('Offset')
    try:
        if offset.node_tree.name == ng_name:
            return offset
    except:
        pass
    try:
        nodes.remove(offset)
    except:
        pass    
    offset = nodes.new(type = 'GeometryNodeGroup')
    ng = node_groups.get(ng_name)
    if ng is None:
        ng = load_ng(ng_dir, ng_name)
    offset.node_tree = ng
    offset.name = 'Offset'
    return offset

####################################### Drivers Utils ######################################

# Add property variable for the driver
def add_prop_var(driver, name, id_type, id, path):
    var = driver.driver.variables.new()
    var.name = name
    var.type = 'SINGLE_PROP'
    var.targets[0].id_type = id_type
    var.targets[0].id = id
    var.targets[0].data_path = path
    
# Convert the data-path to string
def prop_to_string(item, fx, prop):
    s = 'animax_items.coll["'+item+'"].fx["'+fx+'"].'+prop
    return s

# Add Driver
def add_driver(scn, item, fx, attr, name, prop, array_length):
    length = 1 if array_length == 0 else array_length
    for i in range(length):
        driver = attr.driver_add(name, -1 if array_length == 0 else i)
        idx = '' if array_length == 0 else '['+str(i)+']'
        path = prop_to_string(item.name, fx.name, prop+idx)
        add_prop_var(driver, 'prop', 'SCENE', scn, path)
        driver.driver.expression = 'prop'
    
############################################################################################
######################################## Nodes #############################################
############################################################################################

# Transform Effect Node
def transform_fx_node(context, nodes, links, item, fx, ng_dir):
    scn = context.scene
    ng_name = 'ANIMAX_Transform_Effect_NG'
    fx_node = get_fx_node(nodes, ng_name)
    if fx_node is not None:
        return fx_node
    fx_node = add_fx_node(nodes, links, fx, ng_name, ng_dir)
    
    ################# Drivers ########################
    # Location
    add_driver(scn, item, fx, fx_node.inputs[6], "default_value", 'location', 3)
    # Random Location
    add_driver(scn, item, fx, fx_node.inputs[9], "default_value", 'random_location', 3)
    # Rotation
    add_driver(scn, item, fx, fx_node.inputs[7], "default_value", 'rotation', 3)
    # Random Rotation
    add_driver(scn, item, fx, fx_node.inputs[11], "default_value", 'random_rotation', 3)
    # Scale
    add_driver(scn, item, fx, fx_node.inputs[8], "default_value", 'scale', 3)
    # Random Scale
    add_driver(scn, item, fx, fx_node.inputs[13], "default_value", 'random_scale', 3)
    # Explode Location
    add_driver(scn, item, fx, fx_node.inputs[1], "default_value", 'explode_location', 0)
    # Random Location Seed
    add_driver(scn, item, fx, fx_node.inputs[10], "default_value", 'random_location_seed', 0)
    # Random Rotation
    add_driver(scn, item, fx, fx_node.inputs[12], "default_value", 'random_rotation_seed', 0)
    # Random Scale
    add_driver(scn, item, fx, fx_node.inputs[14], "default_value", 'random_scale_seed', 0)
    # Replace Location
    add_driver(scn, item, fx, fx_node.inputs[15], "default_value", 'replace_location', 0)
    # Replace Rotation
    add_driver(scn, item, fx, fx_node.inputs[16], "default_value", 'replace_rotation', 0)
    # Replace Scale
    add_driver(scn, item, fx, fx_node.inputs[17], "default_value", 'replace_scale', 0)
    # Uniform Scale
    add_driver(scn, item, fx, fx_node.inputs[18], "default_value", 'uniform_scale', 0)
    # Uniform Scale Value
    add_driver(scn, item, fx, fx_node.inputs[19], "default_value", 'uniform_scale_value', 0)
    # Uniform Random Scale
    add_driver(scn, item, fx, fx_node.inputs[20], "default_value", 'uniform_random_scale', 0)
    # Fade Out
    add_driver(scn, item, fx, fx_node.inputs[21], "default_value", 'scale_fade_out', 0)
    # Fade Out Smooth Step
    add_driver(scn, item, fx, fx_node.inputs[22], "default_value", 'fade_out_smooth_step', 0)
    return fx_node

# Noise Effect Node
def noise_fx_node(context, nodes, links, item, fx, ng_dir):
    scn = context.scene
    ng_name = 'ANIMAX_Noise_Effect_NG'
    fx_node = get_fx_node(nodes, ng_name)
    if fx_node is not None:
        return fx_node
    fx_node = add_fx_node(nodes, links, fx, ng_name, ng_dir)
    
    ################# Additional Nodes ###################
    # Curve Node
    curve = nodes.new(type = 'ShaderNodeFloatCurve')
    curve.name = 'Rotation_Curve'
    curve.location[0] = x_margine*4
    # Align Euler Node
    align_euler = nodes.new(type = 'FunctionNodeAlignEulerToVector')
    align_euler.name = 'Align_Euler'
    align_euler.location[0] = x_margine*5
    
    ######################## Links ########################
    links.new(nodes['Group Input'].outputs[0], curve.inputs[1])
    links.new(curve.outputs[0], align_euler.inputs[1])
    links.new(fx_node.outputs[1], align_euler.inputs[0])
    links.new(fx_node.outputs[3], align_euler.inputs[2])
    
    ################# Drivers ########################
    # Align axis
    add_driver(scn, item, fx, align_euler, "axis", 'axis_enum', 0)
    
    # Location
    add_driver(scn, item, fx, fx_node.inputs[4], "default_value", 'location', 3)
    # Rotation
    add_driver(scn, item, fx, fx_node.inputs[5], "default_value", 'rotation', 3)
    # Scale
    add_driver(scn, item, fx, fx_node.inputs[6], "default_value", 'scale', 3)
    # Frequency
    add_driver(scn, item, fx, fx_node.inputs[10], "default_value", 'noise_frequency', 0)
    # Scale
    add_driver(scn, item, fx, fx_node.inputs[11], "default_value", 'noise_scale', 0)
    # Detail
    add_driver(scn, item, fx, fx_node.inputs[12], "default_value", 'noise_detail', 0)
    # Roughness
    add_driver(scn, item, fx, fx_node.inputs[13], "default_value", 'noise_roughness', 0)
    # Distortion
    add_driver(scn, item, fx, fx_node.inputs[14], "default_value", 'noise_distortion', 0)
    # Random Seed
    add_driver(scn, item, fx, fx_node.inputs[15], "default_value", 'random_seed', 0)
    # Uniform Scale
    add_driver(scn, item, fx, fx_node.inputs[16], "default_value", 'uniform_scale', 0)
    # Uniform Scale Value
    add_driver(scn, item, fx, fx_node.inputs[17], "default_value", 'uniform_scale_value', 0)
    # Continuous Animation
    add_driver(scn, item, fx, fx_node.inputs[18], "default_value", 'continuous_animation', 0)
    # Fade Out
    add_driver(scn, item, fx, fx_node.inputs[19], "default_value", 'scale_fade_out', 0)
    # Fade Out Smooth Step
    add_driver(scn, item, fx, fx_node.inputs[20], "default_value", 'fade_out_smooth_step', 0)
    return fx_node

# Follow Curve Effect Node
def follow_curve_fx_node(context, nodes, links, item, fx, ng_dir):
    scn = context.scene
    ng_name = 'ANIMAX_Follow_Curve_Effect_NG'
    fx_node = get_fx_node(nodes, ng_name)
    if fx_node is not None:
        return fx_node
    fx_node = add_fx_node(nodes, links, fx, ng_name, ng_dir)
    
    ################# Additional Nodes ###################
    # Curve Node
    curve = nodes.new(type = 'ShaderNodeFloatCurve')
    curve.name = 'Rotation_Curve'
    curve.location[0] = x_margine*4
    # Align Euler Node
    align_euler = nodes.new(type = 'FunctionNodeAlignEulerToVector')
    align_euler.name = 'Align_Euler'
    align_euler.location[0] = x_margine*5
    
    ######################## Links ########################
    links.new(nodes['Group Input'].outputs[0], curve.inputs[1])
    links.new(curve.outputs[0], align_euler.inputs[1])
    links.new(fx_node.outputs[1], align_euler.inputs[0])
    links.new(fx_node.outputs[3], align_euler.inputs[2])
    
    ##################### Drivers ##########################
    # Align axis
    add_driver(scn, item, fx, align_euler, "axis", 'axis_enum', 0)
    
    # World position
    add_driver(scn, item, fx, fx_node.inputs[6], "default_value", 'follow_world_position', 0)
    # Reverse Curve
    add_driver(scn, item, fx, fx_node.inputs[7], "default_value", 'follow_reverse_curve', 0)
    # Curve Resolution
    add_driver(scn, item, fx, fx_node.inputs[8], "default_value", 'follow_resolution', 0)
    # Invert Axis
    add_driver(scn, item, fx, fx_node.inputs[9], "default_value", 'invert_axis', 0)
    # Number of turns
    add_driver(scn, item, fx, fx_node.inputs[10], "default_value", 'num_turns', 0)
    # Fade Out
    add_driver(scn, item, fx, fx_node.inputs[11], "default_value", 'scale_fade_out', 0)
    # Fade Out Smooth Step
    add_driver(scn, item, fx, fx_node.inputs[12], "default_value", 'fade_out_smooth_step', 0)
    return fx_node

# Attract Effect Node
def attract_fx_node(context, nodes, links, item, fx, ng_dir):
    scn = context.scene
    ng_name = 'ANIMAX_Attract_Effect_NG'
    fx_node = get_fx_node(nodes, ng_name)
    if fx_node is not None:
        return fx_node
    fx_node = add_fx_node(nodes, links, fx, ng_name, ng_dir)
    
    ################# Additional Nodes ###################
    # Curve Node
    curve = nodes.new(type = 'ShaderNodeFloatCurve')
    curve.name = 'Rotation_Curve'
    curve.location[0] = x_margine*4
    # Align Euler Node
    align_euler = nodes.new(type = 'FunctionNodeAlignEulerToVector')
    align_euler.name = 'Align_Euler'
    align_euler.location[0] = x_margine*5
    
    ######################## Links ########################
    links.new(nodes['Group Input'].outputs[0], curve.inputs[1])
    links.new(curve.outputs[0], align_euler.inputs[1])
    links.new(fx_node.outputs[1], align_euler.inputs[0])
    links.new(fx_node.outputs[3], align_euler.inputs[2])
    
    ##################### Drivers ##########################
    # Align axis
    add_driver(scn, item, fx, align_euler, "axis", 'axis_enum', 0)
    
    # Factor
    add_driver(scn, item, fx, fx_node.inputs[6], "default_value", 'factor', 0)
    # Collapse
    add_driver(scn, item, fx, fx_node.inputs[7], "default_value", 'collapse_axis', 3)
    # Invert Axis
    add_driver(scn, item, fx, fx_node.inputs[9], "default_value", 'invert_axis', 0)
    # Fade Out
    add_driver(scn, item, fx, fx_node.inputs[10], "default_value", 'scale_fade_out', 0)
    # Fade Out Smooth Step
    add_driver(scn, item, fx, fx_node.inputs[11], "default_value", 'fade_out_smooth_step', 0)
    return fx_node

# Explode Effect Node
def explode_fx_node(context, nodes, links, item, fx, ng_dir):
    scn = context.scene
    ng_name = 'ANIMAX_Explode_Effect_NG'
    fx_node = get_fx_node(nodes, ng_name)
    if fx_node is not None:
        return fx_node
    fx_node = add_fx_node(nodes, links, fx, ng_name, ng_dir)
    
    ##################### Drivers ##########################
    # Velocity
    add_driver(scn, item, fx, fx_node.inputs[4], "default_value", 'velocity', 3)
    # Randomize
    add_driver(scn, item, fx, fx_node.inputs[5], "default_value", 'random_location', 3)
    # Gravity
    add_driver(scn, item, fx, fx_node.inputs[6], "default_value", 'gravity', 3)
    # Time Multiplier
    add_driver(scn, item, fx, fx_node.inputs[7], "default_value", 'time_multiplier', 0)
    # Random Seed
    add_driver(scn, item, fx, fx_node.inputs[8], "default_value", 'random_location_seed', 0)
    # Fade Out
    add_driver(scn, item, fx, fx_node.inputs[9], "default_value", 'scale_fade_out', 0)
    # Fade Out Smooth Step
    add_driver(scn, item, fx, fx_node.inputs[10], "default_value", 'fade_out_smooth_step', 0)
    return fx_node

# Repulse Effect Node
def repulse_fx_node(context, nodes, links, item, fx, ng_dir):
    scn = context.scene
    ng_name = 'ANIMAX_Repulse_Effect_NG'
    fx_node = get_fx_node(nodes, ng_name)
    if fx_node is not None:
        return fx_node
    fx_node = add_fx_node(nodes, links, fx, ng_name, ng_dir)
    
    ##################### Drivers ##########################
    # Force
    add_driver(scn, item, fx, fx_node.inputs[5], "default_value", 'force', 3)
    # Distance
    add_driver(scn, item, fx, fx_node.inputs[6], "default_value", 'distance', 0)
    # Method
    add_driver(scn, item, fx, fx_node.inputs[7], "default_value", 'repulse_method', 0)
    # Fade Out
    add_driver(scn, item, fx, fx_node.inputs[9], "default_value", 'scale_fade_out', 0)
    # Fade Out Smooth Step
    add_driver(scn, item, fx, fx_node.inputs[10], "default_value", 'fade_out_smooth_step', 0)
    return fx_node

# Project Effect Node
def project_fx_node(context, nodes, links, item, fx, ng_dir):
    scn = context.scene
    ng_name = 'ANIMAX_Project_Effect_NG'
    fx_node = get_fx_node(nodes, ng_name)
    if fx_node is not None:
        return fx_node
    fx_node = add_fx_node(nodes, links, fx, ng_name, ng_dir)
    
    ################# Additional Nodes ###################
    # Curve Node
    curve = nodes.new(type = 'ShaderNodeFloatCurve')
    curve.name = 'Rotation_Curve'
    curve.location[0] = x_margine*4
    # Align Euler Node
    align_euler = nodes.new(type = 'FunctionNodeAlignEulerToVector')
    align_euler.name = 'Align_Euler'
    align_euler.location[0] = x_margine*5
    
    ######################## Links ########################
    links.new(nodes['Group Input'].outputs[0], curve.inputs[1])
    links.new(curve.outputs[0], align_euler.inputs[1])
    links.new(fx_node.outputs[1], align_euler.inputs[0])
    links.new(fx_node.outputs[3], align_euler.inputs[2])
    
    ##################### Drivers ##########################
    # Align axis
    add_driver(scn, item, fx, align_euler, "axis", 'axis_enum', 0)
    
    # Ray direction
    add_driver(scn, item, fx, fx_node.inputs[9], "default_value", 'ray_direction', 3)
    # Factor
    add_driver(scn, item, fx, fx_node.inputs[5], "default_value", 'factor', 0)
    # Switch
    add_driver(scn, item, fx, fx_node.inputs[6], "default_value", 'col_ob_switch', 0)
    # Ray length
    add_driver(scn, item, fx, fx_node.inputs[10], "default_value", 'ray_length', 0)
    # Fade Out Missed
    add_driver(scn, item, fx, fx_node.inputs[11], "default_value", 'fade_out_missed', 0)
    # Fade Out
    add_driver(scn, item, fx, fx_node.inputs[12], "default_value", 'scale_fade_out', 0)
    # Fade Out Smooth Step
    add_driver(scn, item, fx, fx_node.inputs[13], "default_value", 'fade_out_smooth_step', 0)
    return fx_node

# Collapse Effect Node
def collapse_fx_node(context, nodes, links, item, fx, ng_dir):
    scn = context.scene
    ng_name = 'ANIMAX_Collapse_Effect_NG'
    fx_node = get_fx_node(nodes, ng_name)
    if fx_node is not None:
        return fx_node
    fx_node = add_fx_node(nodes, links, fx, ng_name, ng_dir)
    
    ##################### Drivers ##########################
    # Collapse Axis
    add_driver(scn, item, fx, fx_node.inputs[6], "default_value", 'collapse_axis', 3)
    # Collapse Offset
    add_driver(scn, item, fx, fx_node.inputs[7], "default_value", 'collapse_offset', 3)
    # Collapse Rotation
    add_driver(scn, item, fx, fx_node.inputs[8], "default_value", 'collapse_rotation', 3)
    # Fade Out
    add_driver(scn, item, fx, fx_node.inputs[9], "default_value", 'scale_fade_out', 0)
    # Fade Out Smooth Step
    add_driver(scn, item, fx, fx_node.inputs[10], "default_value", 'fade_out_smooth_step', 0)
    return fx_node

# Wave Effect Node
def wave_fx_node(context, nodes, links, item, fx, ng_dir):
    scn = context.scene
    ng_name = 'ANIMAX_Wave_Effect_NG'
    fx_node = get_fx_node(nodes, ng_name)
    if fx_node is not None:
        return fx_node
    fx_node = add_fx_node(nodes, links, fx, ng_name, ng_dir)
    fx.noise_scale = 0.1
    fx.noise_frequency = 5.0
    
    ##################### Drivers ##########################
    # Coordinates Rotation
    add_driver(scn, item, fx, fx_node.inputs[7], "default_value", 'coord_rotation', 3)
    # Location
    add_driver(scn, item, fx, fx_node.inputs[8], "default_value", 'location', 3)
    # Rotation
    add_driver(scn, item, fx, fx_node.inputs[9], "default_value", 'rotation', 3)
    # Scale
    add_driver(scn, item, fx, fx_node.inputs[10], "default_value", 'scale', 3)
    # Frequency
    add_driver(scn, item, fx, fx_node.inputs[11], "default_value", 'noise_frequency', 0)
    # Scale
    add_driver(scn, item, fx, fx_node.inputs[12], "default_value", 'noise_scale', 0)
    # Detail
    add_driver(scn, item, fx, fx_node.inputs[14], "default_value", 'noise_detail', 0)
    # Roughness
    add_driver(scn, item, fx, fx_node.inputs[16], "default_value", 'noise_roughness', 0)
    # Distortion
    add_driver(scn, item, fx, fx_node.inputs[13], "default_value", 'noise_distortion', 0)
    # Details Scale
    add_driver(scn, item, fx, fx_node.inputs[15], "default_value", 'details_scale', 0)
    # Phase Offset
    add_driver(scn, item, fx, fx_node.inputs[17], "default_value", 'wave_phase', 0)
    # Continuous Animation
    add_driver(scn, item, fx, fx_node.inputs[18], "default_value", 'continuous_animation', 0)
    # Fade Out
    add_driver(scn, item, fx, fx_node.inputs[19], "default_value", 'scale_fade_out', 0)
    # Fade Out Smooth Step
    add_driver(scn, item, fx, fx_node.inputs[20], "default_value", 'fade_out_smooth_step', 0)
    return fx_node

# Orbit Effect Node
def orbit_fx_node(context, nodes, links, item, fx, ng_dir):
    scn = context.scene
    ng_name = 'ANIMAX_Orbit_Effect_NG'
    fx_node = get_fx_node(nodes, ng_name)
    if fx_node is not None:
        return fx_node
    fx_node = add_fx_node(nodes, links, fx, ng_name, ng_dir)
    
    ##################### Drivers ##########################
    # Rotation
    add_driver(scn, item, fx, fx_node.inputs[6], "default_value", 'rotation', 3)
    # Fade Out
    add_driver(scn, item, fx, fx_node.inputs[7], "default_value", 'scale_fade_out', 0)
    # Fade Out Smooth Step
    add_driver(scn, item, fx, fx_node.inputs[8], "default_value", 'fade_out_smooth_step', 0)
    return fx_node

# Robotic Effect Node
def robotic_fx_node(context, nodes, links, item, fx, ng_dir):
    scn = context.scene
    ng_name = 'ANIMAX_Robotic_Effect_NG'
    fx_node = get_fx_node(nodes, ng_name)
    if fx_node is not None:
        return fx_node
    fx_node = add_fx_node(nodes, links, fx, ng_name, ng_dir)
    
    ##################### Drivers ##########################
    # Distance
    add_driver(scn, item, fx, fx_node.inputs[5], "default_value", 'distance_vec', 3)
    # Rotation
    add_driver(scn, item, fx, fx_node.inputs[6], "default_value", 'rotation', 3)
    # Random Seed
    add_driver(scn, item, fx, fx_node.inputs[7], "default_value", 'random_seed', 0)
    # Fade Out
    add_driver(scn, item, fx, fx_node.inputs[8], "default_value", 'scale_fade_out', 0)
    # Fade Out Smooth Step
    add_driver(scn, item, fx, fx_node.inputs[9], "default_value", 'fade_out_smooth_step', 0)
    return fx_node

# Empty Effect Node
def empty_fx_node(context, nodes, links, item, fx, ng_dir):
    scn = context.scene
    ng_name = 'ANIMAX_Empty_Effect_NG'
    fx_node = get_fx_node(nodes, ng_name)
    if fx_node is not None:
        return fx_node
    fx_node = add_fx_node(nodes, links, fx, ng_name, ng_dir)
    
    ##################### Drivers ##########################
    # Fade Out
    add_driver(scn, item, fx, fx_node.inputs[4], "default_value", 'scale_fade_out', 0)
    # Fade Out Smooth Step
    add_driver(scn, item, fx, fx_node.inputs[5], "default_value", 'fade_out_smooth_step', 0)
    return fx_node

# Scene Time Node
def scene_time_node(nodes):
    scene_time = nodes.get('ANIMAX_Time')
    if scene_time is None:
        scene_time = nodes.new(type = 'GeometryNodeInputSceneTime')
        scene_time.name = 'ANIMAX_Time'
    return scene_time

# Collection Info Node
def coll_info_node(nodes):
    coll_info = nodes.get('ANIMAX_Collection')
    if coll_info is None:
        coll_info = nodes.new(type = 'GeometryNodeCollectionInfo')
        coll_info.inputs[1].default_value = True
        coll_info.name = 'ANIMAX_Collection'
    return coll_info

# Text Input Node
def text_input_node(nodes):
    text = nodes.get('ANIMAX_Text')
    if text is None:
        text = nodes.new(type = 'GeometryNodeStringToCurves')
        text.inputs[0].default_value = 'Text'
        text.name = 'ANIMAX_Text'
        text.pivot_mode = 'MIDPOINT'
    return text

# Custom NG Node
def custom_ng_node(nodes, ng_dir, item):
    custom_ng = nodes.get('ANIMAX_Custom_NG')
    if custom_ng is None:
        custom_ng = nodes.new(type = 'GeometryNodeGroup')
        custom_ng.name = 'ANIMAX_Custom_NG'
    node_groups = bpy.data.node_groups
    empty_ng = node_groups.get('ANIMAX_Empty_NG')
    if empty_ng is None:
        empty_ng = load_ng(ng_dir, 'ANIMAX_Empty_NG')
    ng = item.custom_ng
    if ng is None:
        custom_ng.node_tree = empty_ng
        return custom_ng
    items_tree = ng.interface.items_tree
    output_sockets = [i for i in items_tree if i.item_type == 'SOCKET' and i.in_out == 'OUTPUT']
    if not len(output_sockets):
        custom_ng.node_tree = empty_ng
        return custom_ng
    first_output = output_sockets[0]
    if not first_output.socket_type == 'NodeSocketGeometry':
        custom_ng.node_tree = empty_ng
        return custom_ng
    
    try:
        if not custom_ng.node_tree == ng:
            custom_ng.node_tree = ng
    except Exception as e:
        print(str(e))
        custom_ng.node_tree = empty_ng
    return custom_ng

# Fill Curve Node
def fill_curve_node(nodes):
    fill = nodes.get('ANIMAX_Fill_Curve')
    if fill is None:
        fill = nodes.new(type = 'GeometryNodeFillCurve')
        fill.mode = 'NGONS'
        fill.name = 'ANIMAX_Fill_Curve'
    return fill

# Collection Input Node
def coll_input_node(nodes, ng_dir):
    coll_input = nodes.get('ANIMAX_Collection_Input_NG')
    if coll_input is None:
        node_groups = bpy.data.node_groups
        ng = node_groups.get('ANIMAX_Collection_Input_NG')
        if ng is None:
            ng = load_ng(ng_dir, 'ANIMAX_Collection_Input_NG')
        coll_input = nodes.new(type = 'GeometryNodeGroup')
        coll_input.node_tree = ng
        coll_input.name = 'ANIMAX_Collection_Input_NG'
    return coll_input

# Apply Transform Node
def apply_trans_node(nodes, ng_dir, item, scn):
    apply_trans = nodes.get('ANIMAX_Apply_Transform_NG')
    if apply_trans is None:
        node_groups = bpy.data.node_groups
        ng = node_groups.get('ANIMAX_Apply_Transform_NG')
        if ng is None:
            ng = load_ng(ng_dir, 'ANIMAX_Apply_Transform_NG')
        apply_trans = nodes.new(type = 'GeometryNodeGroup')
        apply_trans.node_tree = ng
        apply_trans.name = 'ANIMAX_Apply_Transform_NG'
        apply_trans.inputs[3].default_value = (1.0, 1.0, 1.0)
        #drivers
        # Use Local Location
        driver = apply_trans.inputs[5].driver_add('default_value')
        path = 'animax_items.coll["'+item+'"].use_local_location'
        add_prop_var(driver, 'prop', 'SCENE', scn, path)
        driver.driver.expression = 'prop'
        # Use Local Rotation
        driver = apply_trans.inputs[6].driver_add('default_value')
        path = 'animax_items.coll["'+item+'"].use_local_rotation'
        add_prop_var(driver, 'prop', 'SCENE', scn, path)
        driver.driver.expression = 'prop'
        # Use Local Scale
        driver = apply_trans.inputs[7].driver_add('default_value')
        path = 'animax_items.coll["'+item+'"].use_local_scale'
        add_prop_var(driver, 'prop', 'SCENE', scn, path)
        driver.driver.expression = 'prop'
        # Mute
        driver = apply_trans.inputs[8].driver_add('default_value')
        path = 'animax_items.coll["'+item+'"].fx_enabled'
        add_prop_var(driver, 'prop', 'SCENE', scn, path)
        driver.driver.expression = 'not prop'
        # Delete Invisible
        driver = apply_trans.inputs[9].driver_add('default_value')
        path = 'animax_items.coll["'+item+'"].delete_invisible'
        add_prop_var(driver, 'prop', 'SCENE', scn, path)
        driver.driver.expression = 'prop'
        
    return apply_trans

# Realize Instances Node
def realize_node(nodes, item, ng_dir, scn):
    realize = nodes.get('ANIMAX_Realize_Instances_NG')
    if realize is None:
        node_groups = bpy.data.node_groups
        ng = node_groups.get('ANIMAX_Realize_Instances_NG')
        if ng is None:
            ng = load_ng(ng_dir, 'ANIMAX_Realize_Instances_NG')
        realize = nodes.new(type = 'GeometryNodeGroup')
        realize.node_tree = ng
        realize.name = 'ANIMAX_Realize_Instances_NG'
        #drivers
        driver = realize.inputs[1].driver_add('default_value')
        path = 'animax_items.coll["'+item+'"].realize_instances'
        add_prop_var(driver, 'prop', 'SCENE', scn, path)
        driver.driver.expression = 'not prop'
    return realize

# Extrude Node
def extrude_node(nodes, item, ng_dir, scn):
    extrude = nodes.get('ANIMAX_Extrude_NG')
    if extrude is None:
        node_groups = bpy.data.node_groups
        ng = node_groups.get('ANIMAX_Extrude_NG')
        if ng is None:
            ng = load_ng(ng_dir, 'ANIMAX_Extrude_NG')
        extrude = nodes.new(type = 'GeometryNodeGroup')
        extrude.node_tree = ng
        extrude.name = 'ANIMAX_Extrude_NG'
        # Drivers
        # Merge by distance
        driver = extrude.inputs[2].driver_add('default_value')
        path = 'animax_items.coll["'+item+'"].merge_by_distance'
        add_prop_var(driver, 'prop', 'SCENE', scn, path)
        driver.driver.expression = 'prop'
    return extrude

# Material Node
def set_material_node(nodes, ng_dir):
    mat = nodes.get('ANIMAX_Set_Material_NG')
    if mat is None:
        node_groups = bpy.data.node_groups
        ng = node_groups.get('ANIMAX_Set_Material_NG')
        if ng is None:
            ng = load_ng(ng_dir, 'ANIMAX_Set_Material_NG')
        mat = nodes.new(type = 'GeometryNodeGroup')
        mat.node_tree = ng
        mat.name = 'ANIMAX_Set_Material_NG'
        #drivers
        driver = mat.inputs[3].driver_add('default_value')
        driver.driver.use_self = True
        driver.driver.expression = "self.id_data.nodes['ANIMAX_Set_Material_NG'].inputs[2].default_value == None"
    return mat

# Group Output Node
def group_output_node(nodes):
    group_output = nodes.get('ANIMAX_Output')
    if group_output is None:
        group_output = nodes.new('NodeGroupOutput')
        group_output.name = 'ANIMAX_Output'
    return group_output

# Position Offset Node
def position_node(context, item, nodes, fx, ng_dir):
    scn = context.scene
    ng_name = 'ANIMAX_Time_Offset_NG'
    offset = nodes.get('Offset')
    if offset is not None and offset.node_tree.name == ng_name:
        return offset
    else:
        offset = add_offset_node(scn, item, nodes, fx, ng_name, ng_dir)
    
    # Drivers
    # Ping-Pong
    add_driver(scn, item, fx, offset.inputs[14], "default_value", 'ping_pong', 0)
    # Repeat Time
    add_driver(scn, item, fx, offset.inputs[13], "default_value", 'repeat_time', 0)
    # Invert Direction
    add_driver(scn, item, fx, offset.inputs[10], "default_value", 'invert_direction', 0)
    # Falloff
    add_driver(scn, item, fx, offset.inputs[11], "default_value", 'offset_falloff', 0)
    # Randomize Offset
    add_driver(scn, item, fx, offset.inputs[7], "default_value", 'random_offset', 0)
    # Random Offset Seed
    add_driver(scn, item, fx, offset.inputs[8], "default_value", 'random_offset_seed', 0)
    # Invert Time
    add_driver(scn, item, fx, offset.inputs[6], "default_value", 'invert_time', 0)
    # Control Time
    add_driver(scn, item, fx, offset.inputs[5], "default_value", 'control_time', 0)
    # Time
    add_driver(scn, item, fx, offset.inputs[4], "default_value", 'time', 0)
    # Duration
    add_driver(scn, item, fx, offset.inputs[3], "default_value", 'duration', 0)
    # Start Frame
    add_driver(scn, item, fx, offset.inputs[2], "default_value", 'start_frame', 0)
    return offset

# Effector Node
def effector_node(context, item, nodes, fx, ng_dir):
    scn = context.scene
    ng_name = 'ANIMAX_Effector_NG'
    offset = nodes.get('Offset')
    if offset is not None and offset.node_tree.name == ng_name:
        return offset
    else:
        offset = add_offset_node(scn, item, nodes, fx, ng_name, ng_dir)
    
    # Drivers
    # Distance
    add_driver(scn, item, fx, offset.inputs[2], "default_value", 'effector_distance', 0)
    # Falloff
    add_driver(scn, item, fx, offset.inputs[3], "default_value", 'effector_falloff', 0)
    # Use Scale
    add_driver(scn, item, fx, offset.inputs[5], "default_value", 'use_effector_scale', 0)
    # Invert The Offset
    add_driver(scn, item, fx, offset.inputs[6], "default_value", 'invert_direction', 0)
    return offset

# Texture Offset Node
def texture_node(context, item, nodes, fx, ng_dir):
    scn = context.scene
    texture_type_dict = {
        'LINEAR': 'ANIMAX_Gradient_Texture_NG',
        'RADIAL': 'ANIMAX_Radial_Texture_NG',
        'NOISE': 'ANIMAX_Noise_Texture_NG',
        'IMAGE': 'ANIMAX_Image_Texture_NG',
    }
    ng_name = texture_type_dict[fx.texture_type]
    offset = nodes.get('Offset')
    if offset is not None and offset.node_tree.name == ng_name:
        return offset
    else:
        offset = add_offset_node(scn, item, nodes, fx, ng_name, ng_dir)
    
    # Drivers   
    # Start Frame
    add_driver(scn, item, fx, offset.inputs[2], "default_value", 'start_frame', 0)
    # Duration
    add_driver(scn, item, fx, offset.inputs[3], "default_value", 'duration', 0)
    # Time
    add_driver(scn, item, fx, offset.inputs[4], "default_value", 'time', 0)
    # Control Time
    add_driver(scn, item, fx, offset.inputs[5], "default_value", 'control_time', 0)
    # Invert Time
    add_driver(scn, item, fx, offset.inputs[6], "default_value", 'invert_time', 0)
    # Falloff
    add_driver(scn, item, fx, offset.inputs[7], "default_value", 'offset_falloff', 0)
    # Rotation
    add_driver(scn, item, fx, offset.inputs[9], "default_value", 'texture_rotation', 3)
    # Invert Offset
    add_driver(scn, item, fx, offset.inputs[10], "default_value", 'invert_direction', 0)
    
    if fx.texture_type == 'NOISE':
        # Noise Scale
        add_driver(scn, item, fx, offset.inputs[11], "default_value", 'noise_texture_scale', 0)
        # Noise Detail
        add_driver(scn, item, fx, offset.inputs[12], "default_value", 'noise_texture_detail', 0)
        # Noise Roughness
        add_driver(scn, item, fx, offset.inputs[13], "default_value", 'noise_texture_roughness', 0)
        # Noise Distortion
        add_driver(scn, item, fx, offset.inputs[14], "default_value", 'noise_texture_distortion', 0)
        # Noise Frequency
        add_driver(scn, item, fx, offset.inputs[15], "default_value", 'noise_texture_frequency', 0)
        # Noise Random Seed
        add_driver(scn, item, fx, offset.inputs[16], "default_value", 'random_offset_seed', 0)
    elif fx.texture_type == 'IMAGE':
        # Location
        add_driver(scn, item, fx, offset.inputs[11], "default_value", 'texture_location', 3)
        # Scale
        add_driver(scn, item, fx, offset.inputs[12], "default_value", 'texture_scale', 3)
    return offset


# Easing Node
def easing_node(context, item, nodes, fx, ng_dir):
    scn = context.scene
    node_groups = bpy.data.node_groups
    name = 'Easing'
    if fx.easing_type == 'CURVE':
        easing = nodes.get('Easing_Curve')
        if easing is None:
            easing = nodes.new(type = 'ShaderNodeFloatCurve')
            easing.name = 'Easing_Curve'
            #driver
            add_driver(scn, item, fx, easing.inputs[0], "default_value", 'easing_curve_factor', 0)
        return easing
            
    if fx.easing_type == 'PING_PONG':
        ng_name = 'ANIMAX_Ping_Pong_NG'
    elif fx.easing_type == 'FUNCTION':
        ng_name = easing_functions[fx.easing]
    
    easing = nodes.get('Easing')
    try:
        if easing.node_tree.name == ng_name:
            return easing
    except:
        pass
        
    try:
        nodes.remove(easing)
    except:
        pass
    
    if fx.easing_type == 'FUNCTION' and fx.easing == 'LINEAR':
        return None
    easing = nodes.new(type = 'GeometryNodeGroup')
    ng = node_groups.get(ng_name)
    if ng is None:
        ng = load_ng(ng_dir, ng_name)
    easing.node_tree = ng
    easing.name = 'Easing'
    
    if fx.easing_type == 'FUNCTION':
        if 'BOUNCE' in fx.easing or 'INVERT' in fx.easing:
            return easing
    
    # Drivers
    add_driver(scn, item, fx, easing.inputs[1], "default_value", 'easing_smooth_step', 0)
    if fx.easing_type == 'FUNCTION' and 'ELASTIC' in fx.easing:
        add_driver(scn, item, fx, easing.inputs[2], "default_value", 'easing_frequency', 0)
        add_driver(scn, item, fx, easing.inputs[3], "default_value", 'easing_amplitude', 0)
    
    return easing