# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

import bpy
from bpy.props import BoolProperty, StringProperty
from bpy.types import Panel, Operator


############################################################################################
################################### splitting Utils #########################################
############################################################################################
# Prepare the object to be splited and put it in a new collection
def prepare_ob(context, source_mesh, col_name, keep):
    scn = context.scene
    # Prepare the collection
    collection_name = get_collection_name(col_name)
    collection = bpy.data.collections.new(collection_name)
    scn.collection.children.link(collection)
    # Keep the original object
    if keep:        
        mesh = source_mesh.data.copy()
        ob = source_mesh.copy()
        ob.data = mesh
        source_mesh.hide_set(True)
        source_mesh.hide_render = True
    else:
        ob = source_mesh
        # Unlink from all the collections
        for coll in ob.users_collection:
            try:
                coll.objects.unlink(ob)
            except:
                pass    
        
    if collection.objects.get(ob.name) is None:
        collection.objects.link(ob)
        
    context.view_layer.objects.active = ob
    for obj in scn.objects:
        obj.select_set(False)
    ob.select_set(True)
    
    # Make sure that the object is visible
    ob.hide_viewport = False
    ob.hide_set(False)
    
    # Apply the modifiers if necessary
    if len(ob.modifiers):
        bpy.ops.object.convert(target='MESH')
    
    return ob, collection

# Split the objects using particle system and modifiers
def split_using_particles(context, source_mesh, col_name, keep):
    source_mesh.animax_keep_original = keep
    scn = context.scene
    ob, collection = prepare_ob(context, source_mesh, col_name, keep)
            
    scn.frame_current = 1    
    tri = ob.modifiers.new('ANIMAX TRIANGULATE','TRIANGULATE')
    tri.show_viewport = False
    sub = ob.modifiers.new('ANIMAX SUBSURF','SUBSURF')
    sub.levels = 0
    sub.render_levels = 0
    sub.subdivision_type = 'SIMPLE'
    psys_name = 'ANIMAX PARTICLES'
    psys = ob.modifiers.new(psys_name, 'PARTICLE_SYSTEM')
    psystem = ob.particle_systems[0]
    psystem.name = psys_name
    psystem.settings.count = 100
    psystem.settings.frame_start = 1.0
    psystem.settings.frame_end = 1.0
    psystem.settings.physics_type = 'NO'
    psystem.settings.particle_size = 0.9
    psystem.settings.display_method = 'NONE'
    explode = ob.modifiers.new('ANIMAX EXPLODE','EXPLODE') 
    explode.use_size = True
    solid = ob.modifiers.new('ANIMAX SOLIDIFY','SOLIDIFY') 
    solid.show_viewport = False
    ob['ANIMAX SPLIT'] = 'EDIT'
    ob.animax_collection_name = collection.name
    
# Convert the faces to individual objects
def split_mesh_faces(context, source_mesh, col_name, keep):
    ob, collection = prepare_ob(context, source_mesh, col_name, keep)
    bpy.ops.object.mode_set(mode = 'EDIT')
    bpy.ops.mesh.select_all(action='SELECT')
    bpy.ops.mesh.edge_split()
    bpy.ops.mesh.separate(type='LOOSE')
    bpy.ops.object.mode_set(mode = 'OBJECT')
    bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY')
    
# Convert the loose parts to individual objects
def split_loose_parts(context, source_mesh, col_name, keep):
    ob, collection = prepare_ob(context, source_mesh, col_name, keep)
    bpy.ops.object.mode_set(mode = 'EDIT')
    bpy.ops.mesh.select_all(action='SELECT')
    bpy.ops.mesh.separate(type='LOOSE')
    bpy.ops.object.mode_set(mode = 'OBJECT')
    bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY')
    
# Separate text letters
def separate_letters(context, source_mesh, col_name, keep):
    ob, collection = prepare_ob(context, source_mesh, col_name, keep)
    bpy.ops.object.convert(target='MESH')    
    bpy.ops.object.mode_set(mode = 'EDIT')
    bpy.ops.mesh.select_all(action='SELECT')
    bpy.ops.mesh.remove_doubles()    
    bpy.ops.mesh.separate(type='LOOSE')
    bpy.ops.object.mode_set(mode = 'OBJECT')
    bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY')
    
# Apply the shards (covert the loose parts to seperate objects)    
def apply_shard(scn, ob):
    # Make sure that the object is visible
    ob.hide_viewport = False
    ob.hide_set(False)
    
    for obj in scn.objects:
        if obj != ob:
            obj.select_set(False)
    ob.select_set(True)
    
    if not len(ob.particle_systems):
        return
    
    if 'ANIMAX PARTICLES' not in ob.particle_systems:
        return
    
    try:
        particles = ob.particle_systems['ANIMAX PARTICLES']
        particles.settings.particle_size = 1.0
        if 'ANIMAX TRIANGULATE' in ob.modifiers:
            tri = ob.modifiers['ANIMAX TRIANGULATE']
            if tri.type == 'TRIANGULATE':
                if not tri.show_viewport:
                    bpy.ops.object.modifier_remove(modifier='ANIMAX TRIANGULATE')
        
        if 'ANIMAX SUBSURF' in ob.modifiers:
            subsurf = ob.modifiers['ANIMAX SUBSURF']
            if subsurf.type == 'SUBSURF':
                if not subsurf.levels > 0:
                    bpy.ops.object.modifier_remove(modifier='ANIMAX SUBSURF')            
        
        if 'ANIMAX SOLIDIFY' in ob.modifiers:
            solid = ob.modifiers['ANIMAX SOLIDIFY']
            if solid.type == 'SOLIDIFY':
                if not solid.show_viewport:
                    bpy.ops.object.modifier_remove(modifier='ANIMAX SOLIDIFY')
        
        if 'ANIMAX EXPLODE' in ob.modifiers:
            exp = ob.modifiers['ANIMAX EXPLODE']
            if exp.type == 'EXPLODE':
                exp.use_size = True
                
        bpy.ops.object.convert(target='MESH')
                
        bpy.ops.object.modifier_remove(modifier='ANIMAX PARTICLES')        
        bpy.ops.object.mode_set(mode = 'EDIT')
        bpy.ops.mesh.select_all(action='SELECT')
        bpy.ops.mesh.separate(type='LOOSE')
        bpy.ops.object.mode_set(mode = 'OBJECT')
        bpy.ops.object.shade_smooth_by_angle()
        bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY')                
        
    except Exception as error:
        print(str(error))

# Cancel the Splitting                
def cancel_splitting(scn, ob):
    collections = bpy.data.collections
    if ob.animax_keep_original:
        mesh = ob.data
        coll = scn.collection
        if len(ob.users_collection) > 0:
            coll = ob.users_collection[0]
        coll.objects.unlink(ob)
        bpy.data.meshes.remove(mesh)
        if len(coll.all_objects) == 0:
            bpy.data.collections.remove(coll)
        return
        
    mods = [
        'ANIMAX TRIANGULATE',
        'ANIMAX SUBSURF',
        'ANIMAX EXPLODE',
        'ANIMAX SOLIDIFY',
        'ANIMAX PARTICLES'
    ]
    for mod in mods:
        if mod not in ob.modifiers:
            continue
        bpy.ops.object.modifier_remove(modifier=mod)
    
    
    coll = collections.get(ob.animax_collection_name)
    if coll:
        if coll in ob.users_collection:
            coll.objects.unlink(ob)
        if len(coll.all_objects) == 0:
            bpy.data.collections.remove(coll)
    if not scn.collection.objects.get(ob.name):
        scn.collection.objects.link(ob)
            
    ob['ANIMAX SPLIT'] = 'NONE'
    ob.animax_keep_original = True
    ob.animax_collection_name = 'ANIMAX'
    

# Get collection name
def get_collection_name(name = 'ANIMAX'):
    if not bpy.data.collections.get(name):
        return name
    i = 2
    while bpy.data.collections.get(name + str(i)):
        i += 1
    return name + str(i)
        
############################################################################################
###################################### Operators ###########################################
############################################################################################
# Split the object using Particles system and Modifiers
class ANIMAX_OT_particles_spliting(Operator):
    bl_idname = "animax.particles_spliting"
    bl_label = "Split the object"
    bl_options = {'UNDO'}
    bl_description = "Split the object using Particles system and Modifiers"
    
    col_name : StringProperty(description = 'Name of the new collection')
    dscr = ("If enabled the original object will not be deleted.\n"
        "The newly shattered objects will be created in a new collection")
    keep : BoolProperty(default = True, description = dscr)
    
    @classmethod
    def poll(self, context):
        return context.object is not None
    
    def invoke(self, context, event):
        self.col_name = context.object.name
        return context.window_manager.invoke_props_dialog(self)

    def draw(self, context):
        layout = self.layout
        layout.separator()
        layout.label(text = 'New Collection')
        layout.prop(self, 'col_name', text = '')
        layout.separator()
        layout.prop(self, 'keep', text = 'Keep Original Object')
        layout.separator()
        
    def execute(self, context):
        ob = context.object
        split_using_particles(context, ob, self.col_name, self.keep)
        return {'FINISHED'}
    
# Apply shards (for Particles based splitting)
class ANIMAX_OT_apply_shards(Operator):
    bl_idname = "animax.apply_shards"
    bl_label = "Apply"
    bl_options = {'UNDO'}
    bl_description = "Convert the shards to objects"
        
    def execute(self, context):
        ob = context.object
        if 'ANIMAX SPLIT' in ob:
            if ob['ANIMAX SPLIT'] == 'EDIT':
                ob['ANIMAX SPLIT'] = 'NONE'
                apply_shard(context.scene, ob)
        return {'FINISHED'}
    
# Convert the mesh faces to individual objects
class ANIMAX_OT_split_faces(Operator):
    bl_idname = "animax.split_faces"
    bl_label = "Separate the faces"
    bl_options = {'UNDO'}
    bl_description = "Convert the mesh faces to individual objects"
    
    col_name : StringProperty(description = 'Name of the new collection')
    dscr = ("If enabled the original object will not be deleted.\n"
        "The newly shattered objects will be created in a new collection")
    keep : BoolProperty(default = True, description = dscr)
    
    @classmethod
    def poll(self, context):
        return context.object is not None
    
    def invoke(self, context, event):
        self.col_name = context.object.name
        return context.window_manager.invoke_props_dialog(self)

    def draw(self, context):
        layout = self.layout
        layout.separator()
        layout.label(text = 'New Collection')
        layout.prop(self, 'col_name', text = '')
        layout.separator()
        layout.prop(self, 'keep', text = 'Keep Original Object')
        layout.separator()
        
    def execute(self, context):
        ob = context.object
        split_mesh_faces(context, ob, self.col_name, self.keep)
        return {'FINISHED'}
    
# Convert the loose parts to individual objects
class ANIMAX_OT_split_loose_parts(Operator):
    bl_idname = "animax.split_loose_parts"
    bl_label = "Separate the loose parts"
    bl_options = {'UNDO'}
    bl_description = "Convert the loose parts to individual objects"
    
    col_name : StringProperty(description = 'Name of the new collection')
    dscr = ("If enabled the original object will not be deleted.\n"
        "The newly shattered objects will be created in a new collection")
    keep : BoolProperty(default = True, description = dscr)
    
    @classmethod
    def poll(self, context):
        return context.object is not None
    
    def invoke(self, context, event):
        self.col_name = context.object.name
        return context.window_manager.invoke_props_dialog(self)

    def draw(self, context):
        layout = self.layout
        layout.separator()
        layout.label(text = 'New Collection')
        layout.prop(self, 'col_name', text = '')
        layout.separator()
        layout.prop(self, 'keep', text = 'Keep Original Object')
        layout.separator()
        
    def execute(self, context):
        ob = context.object
        split_loose_parts(context, ob, self.col_name, self.keep)
        return {'FINISHED'} 
    
# Separate the letters
class ANIMAX_OT_separate_letters(Operator):
    bl_idname = "animax.separate_letters"
    bl_label = "Separate the letters"
    bl_options = {'UNDO'}
    bl_description = "Convert the letters to individual objects"
    
    col_name : StringProperty(description = 'Name of the new collection')
    dscr = ("If enabled the original object will not be deleted.\n"
        "The newly shattered objects will be created in a new collection")
    keep : BoolProperty(default = True, description = dscr)
    
    @classmethod
    def poll(self, context):
        return context.object is not None
    
    def invoke(self, context, event):
        self.col_name = context.object.name
        return context.window_manager.invoke_props_dialog(self)

    def draw(self, context):
        layout = self.layout
        layout.separator()
        layout.label(text = 'New Collection')
        layout.prop(self, 'col_name', text = '')
        layout.separator()
        layout.prop(self, 'keep', text = 'Keep Original Object')
        layout.separator()
        
    def execute(self, context):
        ob = context.object
        if ob.type == 'FONT':
            separate_letters(context, ob, self.col_name, self.keep)
        return {'FINISHED'}
    
# Cancel the Splitting process
class ANIMAX_OT_cancel_split(Operator):
    bl_idname = "animax.cancel_splitting"
    bl_label = "Cancel"
    bl_options = {'UNDO'}
    bl_description = "Cancel the Splitting process"
        
    def execute(self, context):
        ob = context.object
        cancel_splitting(context.scene, ob)
        return {'FINISHED'}


############################################################################################
####################################### The UI #############################################
############################################################################################    
# splitting Panel
class ANIMAX_PT_split(Panel):
    bl_label = "Shell Splitting"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_category = "ANIMAX"
    bl_context = "objectmode"        
    bl_parent_id = "ANIMAX_PT_main"
    
    @classmethod
    def poll(self, context):
        if context.scene.animax_tabs != 'FRACTURE':
            return False
        return context.scene.animax_fracture_tabs == 'SPLIT'
        
    def draw(self, context):
        layout = self.layout        
        scn = context.scene
        collections = bpy.data.collections        
        ob = context.object
        col = layout.column()
        
        if not ob:
            col.label(text ='An Active object is required', icon = 'INFO')
            return
        
        if not context.object.type == 'MESH':
            col.label(text ='Only Mesh objects are supported', icon = 'INFO')
            if ob.type == 'FONT':
                col.operator("animax.separate_letters", icon = 'FILE_FONT')
            return
        
        modifiers = ob.modifiers
        col = layout.column()
        
        if 'ANIMAX SPLIT' in ob:
            if ob['ANIMAX SPLIT'] != 'NONE':
                if len(ob.particle_systems) > 0:
                    if 'ANIMAX PARTICLES' in ob.particle_systems:
                        particles = ob.particle_systems['ANIMAX PARTICLES']
                        col.prop(particles, 'seed')
                        col.prop(particles.settings, 'count')
                if 'ANIMAX SUBSURF' in modifiers:
                    subsurf = modifiers['ANIMAX SUBSURF']
                    if subsurf.type == 'SUBSURF':
                        row = col.row(align = True)
                        row.prop(subsurf, 'levels', text = 'Sub divisions')
                        row.prop(subsurf, 'subdivision_type', text = '')
                if 'ANIMAX TRIANGULATE' in modifiers:
                    tri = modifiers['ANIMAX TRIANGULATE']
                    if tri.type == 'TRIANGULATE':
                        col.prop(tri, 'show_viewport', text = 'Triangulate', icon = 'MOD_TRIANGULATE')
                if 'ANIMAX SOLIDIFY' in modifiers:
                    solid = modifiers['ANIMAX SOLIDIFY']
                    if solid.type == 'SOLIDIFY':                            
                        col.prop(solid, 'show_viewport', text = 'Solidify', icon = 'MOD_SOLIDIFY')
                        if solid.show_viewport:
                            col.prop(solid, 'thickness')
                            col.prop(solid, 'material_offset')
                            col.prop(solid, 'material_offset_rim')
                if 'ANIMAX EXPLODE' in modifiers:
                    exp = modifiers['ANIMAX EXPLODE']
                    if exp.type == 'EXPLODE':                            
                        col.prop(exp, 'use_edge_cut', icon = 'MOD_BOOLEAN')
                        col.prop(exp, 'use_size', text = 'Preview shards', icon = 'MOD_BUILD')
                col = layout.column()        
                col = layout.column()        
                row = col.row()        
                row.operator("animax.apply_shards", icon = 'FILE_TICK')        
                row.operator("animax.cancel_splitting", icon = 'CANCEL')
                return
        
        if ob.type == 'MESH':
            col.operator("animax.particles_spliting", icon = 'MOD_BOOLEAN')
            col.separator() 
            col.operator("animax.split_faces", icon = 'MOD_EXPLODE')
            col.separator() 
            col.operator("animax.split_loose_parts", icon = 'MOD_EDGESPLIT')
            
############################################################################################
################################ Register/Unregister #######################################
############################################################################################

classes = (
    ANIMAX_OT_particles_spliting,
    ANIMAX_OT_apply_shards,
    ANIMAX_OT_split_faces,
    ANIMAX_OT_split_loose_parts,
    ANIMAX_OT_separate_letters,
    ANIMAX_OT_cancel_split,
    ANIMAX_PT_split,
)

def register():
    from bpy.utils import register_class
    for cls in classes:
        register_class(cls)
        
    bpy.types.Object.animax_keep_original = BoolProperty(default = True)
    bpy.types.Object.animax_collection_name = StringProperty(
        default = 'ANIMAX',
        description = 'Name of the new collection'
    )

def unregister():
    from bpy.utils import unregister_class
    for cls in reversed(classes):
        unregister_class(cls)

    del bpy.types.Object.animax_keep_original    
    del bpy.types.Object.animax_collection_name

if __name__ == "__main__":
    register()