# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

import bpy, bmesh, time, sys
import numpy as np
from math import dist
from bpy.props import *
from mathutils.kdtree import KDTree
from mathutils.bvhtree import BVHTree
from mathutils import Vector, Matrix
from bpy.types import PropertyGroup
from bpy.types import Panel, Operator

# Get a new unique name for a property
def get_prop_name(name, names):
    new_name = name    
    i = 2
    while new_name in names:
        new_name = name + str(i)
        i += 1
    return new_name

# Print iterations progress
def update_progress(job, progress, start):
    length = 35
    block = int(round(length*progress))
    prog = round(progress*100, 2)
    msg = f'\r{job}: |{"█"*block + "-"*(length-block)}| {prog}%'
    if progress >= 1:
        t = round(time.time() - start, 2)
        sys.stdout.write("\r\033[K")
        msg = f"{job}: DONE in {t} Seconds\n"
    sys.stdout.write(msg)
    sys.stdout.flush()

############################################################################################
#################################### Voronoi Cells #########################################
############################################################################################

# Add a new collection
def new_collection(context, name):
    main_coll = context.scene.collection
    collections = bpy.data.collections
    collection = collections.new(name)
    main_coll.children.link(collection)
    return collection

# Create the bounding box mesh
def create_bound_box(bm, corners):
    # Index of the vertices for every face
    bbox_faces = [
        [0,4,5,1],
        [0,1,2,3],
        [0,4,7,3],
        [4,7,6,5],
        [7,6,2,3],
        [6,2,1,5]
    ]
    # Add the vertices
    verts = [bm.verts.new(v) for v in corners]
    # Layers (geometry attributes)
    neighbour = bm.faces.layers.int.new('neighbour')
    interior = bm.faces.layers.int.new('Interior Faces')
    # Create the bounding box
    for f in bbox_faces:
        vertices = [verts[v] for v in f]
        face = bm.faces.new(vertices)
        face[interior] = 1
        face[neighbour] = -1
    # Recalculate the normals
    bmesh.ops.recalc_face_normals(bm, faces=bm.faces)
    # Add a margin
    for f in bm.faces:
        for v in f.verts:
            v.co += (f.normal*0.1)
    return bm

# Create the cell object
def create_cell_ob(ob, collection, interior_v_group, transfer_normals):
    mesh = bpy.data.meshes.new(name = ob.name + '_Shard')
    obj = bpy.data.objects.new(name = 'Shard', object_data = mesh)
    # add to the collection
    collection.objects.link(obj)
    # copy the materials from the target object
    for mat in ob.data.materials:
        mesh.materials.append(mat)
    # add a vertex group
    if interior_v_group or transfer_normals:
        obj.vertex_groups.new(name="Interior Verts")
    return obj

def intersect_edge_plane(edge, p, n):
    d1, d2 = [(p-v.co).dot(n) for v in edge.verts]
    return not ((d1>0 and d2>0) or (d1<0 and d2<0))

# detect if there is an intersection between the cell and the target
def intersect_mesh_mesh(ob_BVHtree, cell_bm):
    cell_BVHtree = BVHTree.FromBMesh(cell_bm)
    inter = ob_BVHtree.overlap(cell_BVHtree)
    return inter != []

# add a boolean modifier
def add_bool_mod(cell, obj, group):
    mod = cell.modifiers.new(name="Boolean", type='BOOLEAN')
    mod.object = obj
    mod.operation = 'INTERSECT'
    mod.double_threshold = group.overlap_threshold
    mod.use_hole_tolerant = group.use_hole_tolerant
    mod.use_self = group.use_self
    mod.solver = group.bool_solver

# apply the boolean modifier, and other options
def apply_bool_mod(context, cells):
    start = time.time()
    job = 'Applying Boolean'
    sys.stdout.write(f'{job}: ... ')
    sys.stdout.flush()
    group = context.scene.animax_voronoi_fracture
    depsgraph = context.evaluated_depsgraph_get()
    bm = bmesh.new()
    for cell in cells:
        cell_eval = cell.evaluated_get(depsgraph)
        bm.from_mesh(cell_eval.data)
        cell.modifiers.remove(cell.modifiers[-1])
        bm.normal_update()
        bmesh.ops.remove_doubles(bm, verts=bm.verts, dist=0.0001)
        # mark sharp and/or split edges
        sharp_edges = []
        if group.split_sharp or group.sharp_edges or group.transfer_normals:
            for edge in bm.edges:
                if len({face.hide for face in edge.link_faces}) == 2:
                    edge.smooth = False
                    sharp_edges.append(edge)
        if group.split_sharp or group.transfer_normals:
            bm.normal_update()
            bmesh.ops.split_edges(bm, edges=sharp_edges)
        # add interior vertex group
        if group.interior_v_group or group.transfer_normals:
            for vert in bm.verts:
                vert.tag = True
            for face in bm.faces:
                if face.hide:
                    continue
                for bm_vert in face.verts:
                    bm_vert.tag = False
            defvert_lay = bm.verts.layers.deform.verify()
            for vert in bm.verts:
                if vert.tag:
                    vert[defvert_lay][0] = 1.0
        for face in bm.faces:
            face.hide = False
        bm.to_mesh(cell.data)
        bm.clear()
    bm.free()
    t = round(time.time() - start, 2)
    sys.stdout.write("\r\033[K")
    sys.stdout.write(f"{job}: DONE in {t} Seconds\n")
    sys.stdout.flush()
    
# add then apply the data transfer modifier    
def apply_data_transfer_mod(context, cells, ob):
    start = time.time()
    job = 'Transferring Normals'
    sys.stdout.write('f{job}: ... ')
    sys.stdout.flush()
    for cell in cells:
        mod = cell.modifiers.new(name='Data Transfer', type='DATA_TRANSFER')
        mod.object = ob
        mod.vertex_group = "Interior Verts"
        mod.invert_vertex_group = True
        mod.use_loop_data = True
        mod.data_types_loops = {'CUSTOM_NORMAL'}
        mod.loop_mapping = 'POLYINTERP_LNORPROJ'
    deps = context.evaluated_depsgraph_get()
    bm = bmesh.new()
    for cell in cells:
        cell_eval = cell.evaluated_get(deps)
        bm.from_mesh(cell_eval.data)
        cell.modifiers.remove(cell.modifiers[-1])
        bm.to_mesh(cell.data)
        bm.clear()
    bm.free()
    t = str(round(time.time() - start, 2))
    sys.stdout.write("\r\033[K")
    sys.stdout.write(f"{job}: DONE in {t} Seconds\n")
    sys.stdout.flush()
    
# apply the margin, applying it in the first pass is problematic
def apply_margin(i, bms, margin, points, mat_index, smooth):
    neighbour = bms[i].faces.layers.int.get('neighbour')
    neighbours = {f[neighbour] for f in bms[i].faces}
    neighbours.discard(-1)
    for index in neighbours:
        bisect(i, index, bms, points, mat_index, margin, smooth)

# Bisector between two points        
def bisector(idx1, idx2, points, margin):
    p1 = points[idx1]
    p2 = points[idx2]
    l = dist(p1, p2)
    fac = (l-margin)/(2*l)
    plane_co = p1.lerp(p2, fac)
    plane_no = p2 - p1
    plane_no.normalize()
    return plane_co, plane_no

# apply the bisect operation then fill the faces
def bisect(p1, p2, bms, points, mat_index, margin, smooth):
    bm = bms[p1]
    geo = bm.verts[:] + bm.edges[:] + bm.faces[:]
    pos, nor = bisector(p1, p2, points, margin)
    ret = bmesh.ops.bisect_plane(
            bm, geom=geo, dist=0.0001,
            plane_co=pos, plane_no=nor,
            clear_outer=True
        )
    contour = [e for e in ret['geom_cut'] if isinstance(e, bmesh.types.BMEdge)]        
    fill = bmesh.ops.holes_fill(bm, edges=bm.edges)
#    edgnet = bmesh.ops.edgenet_prepare(bm, edges=contour)
#    fill = bmesh.ops.edgeloop_fill(bm, edges=edgnet['edges'])
    neighbour = bm.faces.layers.int.get('neighbour')
    for face in fill['faces']:
        face[neighbour] = p2
        if mat_index != 0:
            face.material_index = mat_index
        face.hide = True
        face.smooth = smooth
    return contour
        
# Insert a new cell
def insert_cell(p, i, kd, bms, points, mat_index, smooth):
    # check if the point is inserted
    def is_inserted(idx):
        return idx < i
    # get the nearest point
    co, nearest_idx, dis = kd.find(p, filter = is_inserted)
    if nearest_idx is None:
        return
    neighbours = {nearest_idx,}
    next_neighbours = {nearest_idx,}
    # find then bisect all the valid neighbouring cells
    while next_neighbours:
        tmp_neighbours = next_neighbours
        next_neighbours = set()
        for idx in tmp_neighbours:
            edges = bisect(i, idx, bms, points, mat_index, 0, smooth)
            if not edges:
                continue
            edges = bisect(idx, i, bms, points, mat_index, 0, smooth)
            neighbour = bms[idx].faces.layers.int.get('neighbour')
            indexs = {f[neighbour] for e in edges for f in e.link_faces}
            next_neighbours.update(indexs - {-1, i} - neighbours)
            neighbours.update(next_neighbours)

# generate the voronoi cells from the bounding box of an object
def generate_cells(context, ob, points, ob_bm):
    s = time.time()
    group = context.scene.animax_voronoi_fracture
    bms = []
    # create a KD tree for the points
    kd = KDTree(len(points))
    for i, co in enumerate(points):
        kd.insert(co, i)
    kd.balance()
    # bmesh for the bounding box
    mat = ob.matrix_world.copy()
    corners = [mat @ Vector(c) for c in ob.bound_box]
    bbox_bm = bmesh.new() 
    create_bound_box(bbox_bm, corners)
    # material index
    mat_index = group.material_index
    if mat_index > len(ob.material_slots)-1:
        mat_index = 0
    # generate the cells (bmesh)
    l = len(points)
    for i, p in enumerate(points):
        bm = bbox_bm.copy()
        bms.append(bm)
        insert_cell(p, i, kd, bms, points, mat_index, group.smooth_interior)
        update_progress("Generating Cells", (i+1)/l, s)
    bbox_bm.free()
    # add the cell objects
    ob_BVHtree = BVHTree.FromBMesh(ob_bm)
    bool_objs = []
    collection = new_collection(context, ob.name + ' Fractured')
    for i, bm in enumerate(bms):
        # apply the margin
        if group.margin > 0:
            apply_margin(i, bms, group.margin, points, mat_index, group.smooth_interior)
        # cleanup before the boolean
        neighbour = bm.faces.layers.int.get('neighbour')
        bm.faces.layers.int.remove(neighbour)
        bmesh.ops.remove_doubles(bm, verts=bm.verts, dist=0.0001)
        bmesh.ops.dissolve_limit(bm, use_dissolve_boundaries=True,
            verts=bm.verts, edges=bm.edges, angle_limit=0.001)
        # add the object
        obj = create_cell_ob(ob, collection, group.interior_v_group, group.transfer_normals)
        # check for intersection, to add the boolean modifier only if necessary
        if intersect_mesh_mesh(ob_BVHtree, bm):
            bool_objs.append(obj)
            add_bool_mod(obj, ob, group)
        bm.to_mesh(obj.data)
        bm.free()
        # preview cells generation (slow)
        #bpy.ops.wm.redraw_timer(type='DRAW_WIN_SWAP', iterations=1)
    del bms[:]
    del kd
    return bool_objs, collection

############################################################################################
################################### Setup Particles ########################################
############################################################################################

# get the resolution of a grid for a number of points
def calc_grid_resolution(count):
    resolution = 2
    while count > resolution ** 3:
        resolution += 1
    return resolution

# Add a particle system
def add_particles_sys(obj, group):
    count = group.points_count
    mod = obj.modifiers.new('ANIMAX_FRACTURE', 'PARTICLE_SYSTEM')
    psystem = obj.particle_systems[-1] 
    psystem.name = 'ANIMAX_FRACTURE'
    psystem.settings.count = count
    psystem.settings.emit_from = 'VOLUME'
    psystem.settings.distribution = group.distribution
    psystem.seed = group.random_seed
    if group.distribution == 'GRID':
         psystem.settings.grid_resolution = calc_grid_resolution(count)
         psystem.settings.grid_random = group.grid_random
    psystem.settings.frame_start = 1.0
    psystem.settings.frame_end = 1.0
    psystem.settings.lifetime = 100000.0
    psystem.settings.use_modifier_stack = True
    psystem.settings.physics_type = 'NO'
    psystem.settings.display_size = 0.025
    psystem.settings.display_method = 'DOT'
    return mod

# get the visibility of the particle systems
def get_psystems_visibility(obj):
    mods = []
    for mod in obj.modifiers:
        if not mod.type == 'PARTICLE_SYSTEM':
            continue
        if not mod.show_viewport:
            continue
        mod.show_viewport = False
        mods.append(mod)
    return mods

############################################################################################
################################### Getting Points #########################################
############################################################################################

# get only the points that their distance to the nearest point is greater than the margin
def valide_margin_points(points, margin):
    kd = KDTree(len(points))
    for i, co in enumerate(points):
        kd.insert(co, i)
    kd.balance()
    def min_dist(index):
        ret = kd.find_range(points[index], margin)
        indexs = []
        if ret:
            indexs = [i[1] for i in ret if i[1] != index]
        return indexs
    removed = []
    for idx, p in enumerate(points):
        ret = min_dist(idx)
        if not ret:
            continue
        ret = [i for i in ret if i not in removed]
        if ret:
            removed.append(idx)
    valid_points = [p for p in points if points.index(p) not in removed]
    return valid_points

# scale the points
def scale_points(points, scale):
    if not len(points):
        return []
    center = sum(points, Vector()) / len(points)
    t = Matrix.Translation(center)
    s = Matrix.Diagonal(scale).to_4x4()
    neg_t = Matrix.Translation(-center)
    mat = t @ s @ neg_t
    points = [mat @ Vector(p) for p in points]
    return points

# get the points inside the bounding box
def points_inside_bbox(points, obj):
    bm = bmesh.new()
    mat = obj.matrix_world.copy()
    corners = [mat @ Vector(c) for c in obj.bound_box]
    bm = bmesh.new() 
    create_bound_box(bm, corners)
    points = points_inside_mesh(points, bm)
    bm.free()
    return points

# get the points inside a mesh
def points_inside_mesh(points, ob_bm):
    ob_BVHtree = BVHTree.FromBMesh(ob_bm)
    directions = [
        Vector((1.0, 0.0, 0.0)),
        Vector((0.0, 1.0, 0.0)),
        Vector((0.0, 0.0, 1.0)),
        Vector((-1.0, 0.0, 0.0)),
        Vector((0.0, -1.0, 0.0)),
        Vector((0.0, 0.0, -1.0))
    ]
    def is_inside(p):
        hits = 0
        for direction in directions:
            co, nor, idx, dis = ob_BVHtree.ray_cast(p, direction)
            if not nor:
                continue
            if nor.dot(direction) > 0.0:
                hits += 1
        return hits > 3
    points = [p for p in points if is_inside(Vector(p))]
    return points

# get the points from a grease pencil object
def points_from_gp(obj):
    layers = [layer for layer in obj.data.layers]
    if not layers:
        return[]
    frames = [frame for layer in layers for frame in layer.frames]
    if not frames:
        return[]
    strokes = [stroke for frame in frames for stroke in frame.strokes]
    if not strokes:
        return[]
    matrix = obj.matrix_world.copy()
    points = [matrix @ p.co.copy() for stroke in strokes for p in stroke.points]
    return points

# get the points from an object    
def points_from_verts(obj, depsgraph):
    ob_eval = obj.evaluated_get(depsgraph)
    if obj.type == 'GPENCIL':
        return points_from_gp(ob_eval)
    
    points = []
    if obj.type == 'MESH':
        mesh = ob_eval.data
    else:
        try:
            mesh = ob_eval.to_mesh()
        except:
            mesh = None

    if mesh is not None:
        matrix = obj.matrix_world.copy()
        points = [matrix @ v.co for v in mesh.vertices]
        ob_eval.to_mesh_clear()
    return points

# get the points from a particle system            
def points_from_particles(obj, depsgraph):
    points = []
    obj_eval = obj.evaluated_get(depsgraph)
    points = [p.location.copy()
              for psys in obj_eval.particle_systems
              for p in psys.particles]
    return points

# get the points from the grease pencil annotations
def points_from_annotations(scn):
    points = []
    gp = scn.grease_pencil
    if not hasattr(gp, 'layers'):
        return []
    layers = [layer for layer in gp.layers]
    frames = [frame for layer in layers for frame in layer.frames]
    if not frames:
        return[]
    strokes = [stroke for frame in frames for stroke in frame.strokes]
    if not strokes:
        return[]
    points = [p.co.copy() for stroke in strokes for i, p in enumerate(stroke.points)]
    return points

# get all the available points
def get_all_points(context, group, ob, eval_ob):
    scn = context.scene
    count = group.points_count
    use_annotations = group.use_annotations
    percentage = group.annotation_percentage
    np.random.seed(group.random_seed)
    # get the points (particles, verts)
    if group.method == 'QUICK':
        modifiers = get_psystems_visibility(ob)
        psystem = add_particles_sys(ob, group)
        depsgraph = context.evaluated_depsgraph_get()
        points = points_from_particles(ob, depsgraph)
        for mod in modifiers:
            mod.show_viewport = True
        ob.modifiers.remove(psystem)
    else:
        points = []
        depsgraph = context.evaluated_depsgraph_get()
        if 'PARTICLES' in group.points_source:
            points.extend(points_from_particles(ob, depsgraph))
        if 'VERTS' in group.points_source:
            collection = group.collection
            if hasattr(collection, 'all_objects'):
                for obj in collection.all_objects:
                    points.extend(points_from_verts(obj, depsgraph))
    # scale the points
    if len(points):
        if group.scale != (1.0, 1.0, 1.0):
            points = scale_points(points, group.scale)
        points = points_inside_bbox(points, eval_ob)
        points = np.unique(points, axis=0)
        np.random.shuffle(points)
    # get the annotation points    
    if use_annotations:
        annotation_points = points_from_annotations(scn)
        if group.scale != (1.0, 1.0, 1.0):
            annotation_points = scale_points(annotation_points, group.scale)
        annotation_points = points_inside_bbox(annotation_points, eval_ob)
        annotation_points = np.unique(annotation_points, axis=0)
        np.random.shuffle(annotation_points)
        # adjust the ratio of points
        if len(annotation_points):
            if percentage == 100:
                points = annotation_points
            else:
                fac = int(len(annotation_points) * percentage / 100)
                if len(points) > len(annotation_points)-fac:
                    points = points[:len(annotation_points)-fac]
                annotation_points = annotation_points[:fac]
                points = np.concatenate((points, annotation_points), axis=0)

    if not len(points):
        return []
            
    points = np.unique(points, axis=0)
    if count < len(points):
        np.random.shuffle(points)
        points = points[:count]
    # sort the points from the center (faster)
    # TODO a more sophisticated sorting algorithm
    center = points.mean(axis = 0)
    points = sorted(points, key = lambda p : dist(p, center))
    # convert to Vector
    points = [Vector(p) for p in points]
    # get only the valid points if the margin is applied
    if group.margin > 0.0:
        points = valide_margin_points(points, group.margin)
    
#    # debug
#    bm = bmesh.new()
#    for p in points:
#        bm.verts.new(p)
#    bm.to_mesh(bpy.data.objects['Cube'].data)
#    bm.free()
    return points

############################################################################################
##################################### Properties ###########################################
############################################################################################
class VoronoiFractureGroup(PropertyGroup):
    expand_panel : BoolVectorProperty(
        name = 'Expand/Collapse',
        default = (True, True, False, False),
        size = 4,
        description = 'Expand/Collapse this panel'
    )
    method: EnumProperty(
        name = 'Method',
        items = (
            ('QUICK', 'Quick', 'The add-on generate the points'),
            ('ADVANCED', 'Advanced', 'The points are taken from a particle system the vertices of an object'),
        ),
        default= 'QUICK',
        description = 'Points generation method'
    )
    points_source: EnumProperty(
        name = 'Points Source',
        items = (
            ('PARTICLES', 'Particles', ''),
            ('VERTS', 'Vertices', ''),
        ),
        options={'ENUM_FLAG'},
        default={'PARTICLES'},
        description = 'Source of points\n'
        'Hold the Shift key select multiple options'
    )
    collection : PointerProperty(
        name = 'Collection',
        type = bpy.types.Collection,
        description = 'Use the vertices of the objects inside this collection as points'
    )
    use_annotations : BoolProperty(
        name = 'Use Annotations',
        default = False,
        description = 'Use the points from the Grease Pencil Annotations'
    )
    annotation_percentage : IntProperty(
        name = 'Percentage',
        default = 50,
        min = 1,
        max = 100,
        subtype = 'PERCENTAGE',
        description = 'The ratio of the annotation points'
    )
    max : IntProperty(
        name = 'Max',
        default = 1000,
        min = 0,
        max = 25000,
        description = 'The maximum number of points.\n'
        'If set to 0 all the points will be included'
    )
    random_seed : IntProperty(
        name = 'Random Seed',
        default = 0,
        min = 0,
        max = 100000,
        description = 'Random Seed'
    )
    randomize_seed : BoolProperty(
        name = 'Randomize Seed',
        default = True,
        description = 'Set a different random seed for every object'
    )
    points_count : IntProperty(
        name = 'Count',
        default = 100,
        min = 2,
        max = 25000,
        description = 'Max number of points to generate the voronoi cells'
    )
    material_index : IntProperty(
        name = 'Material Index',
        default = 0,
        min = 0,
        max = 1000,
        description = 'Index of the Material for the interior faces'
    )
    distribution : EnumProperty(
        name = 'Distribution',
        items = (
            ('GRID', 'Grid', ''),
            ('RAND', 'Random', ''),
        ),
        default = 'RAND',
        description = 'How to distribute the points'
    )
    grid_random : FloatProperty(
        name = 'Random',
        default = 0.5,
        min = 0,
        max = 1.0,
        description = 'Randomomize the position of the points'
    )
    margin : FloatProperty(
        name = 'Margin',
        default = 0.0,
        min = 0.0,
        max = 1000.0,
        description = 'Create a gap between the shards'
    )
    scale : FloatVectorProperty(
        name = 'Scale',
        default = (1.0, 1.0, 1.0),
        subtype = 'XYZ',
        description = 'Scale of the shards'
    )
    interior_v_group : BoolProperty(
        name = 'Interior Vertex Group',
        default = False,
        description = 'Create a new Vertex Group for the interior vertices'
    )
    sharp_edges : BoolProperty(
        name = 'Sharp Edges',
        default = False,
        description = 'The edges between the interior and exterior faces will be sharp'
    )
    split_sharp : BoolProperty(
        name = 'Split Sharp Edges',
        default = False,
        description = 'Split the sharp edges'
    )
    smooth_interior : BoolProperty(
        name = 'Smooth Interior',
        default = False,
        description = 'Smooth interior faces'
    )
    transfer_normals : BoolProperty(
        name = 'Transfer Normals',
        default = False,
        description = 'Transfer the normals from the target objet to the shards'
    )
    bool_solver : EnumProperty(
        name = 'Boolean Solver',
        items = (
            ('FAST', 'Fast', ''),
            ('EXACT', 'Exact', ''),
        ),
        default = 'EXACT',
        description = 'Method for calculating booleans'
    )
    overlap_threshold : FloatProperty(
        name = 'Overlap Threshold',
        default = 0.0,
        min = 0.0,
        max = 1.0,
        description = 'Threshold for checking overlapping geometry'
    )
    use_self : BoolProperty(
        name = 'Self Intersection',
        default = False,
        description = 'Allow self intersection'
    )
    use_hole_tolerant : BoolProperty(
        name = 'Hole Tolerant',
        default = False,
        description = 'Better results when there are holes (Slower)'
    )

############################################################################################
###################################### Operators ###########################################
############################################################################################
# Fracture
class ANIMAX_OT_voronoi_fracture(Operator):
    bl_idname = "animax.voronoi_fracture"
    bl_label = "Fracture"
    bl_options = {'UNDO'}
    bl_description = "Start the fracturing operation"
    
    @classmethod
    def poll(cls, context):
        if not hasattr(context.scene, 'animax_voronoi_fracture'):
            return False
        return context.selected_editable_objects != []
    
    def execute(self, context):
        objects = context.selected_editable_objects[:]
        group = context.scene.animax_voronoi_fracture
        random_seed = group.random_seed
        add_interior_v_group = group.interior_v_group
        for ob in objects:
            s = time.time()
            if ob.type != 'MESH':
                self.report({'WARNING'}, 'Only Mesh Objects Supported')
                continue
            # unhide the faces
            faces = ob.data.polygons
            faces.foreach_set('hide', [False,]*len(faces))
            # get the points
            depsgraph = context.evaluated_depsgraph_get()
            eval_ob = ob.evaluated_get(depsgraph)
            points = get_all_points(context, group, ob, eval_ob)
            if len(points) < 2:
                self.report({'WARNING'}, 'Not enough points to generate a Voronoi Diagram')
                continue
            # set the random seed
            if group.randomize_seed and len(objects):
                group.random_seed += 1
            # get the points inside the mesh
            ob_bm = bmesh.new()
            ob_bm.from_mesh(eval_ob.data)
            ob_bm.transform(ob.matrix_world)
            points_inside = points_inside_mesh(points, ob_bm)
            # hide the target object
            ob.hide_set(True)
            ob.hide_render = True
            # start the fracturing
            print(f'{"#"*31} Fracturing {"#"*31}')
            print('Points Found:', len(points))
            # generate the cells
            bool_cells, collection = generate_cells(context, ob, points, ob_bm)
            ob_bm.free()
            # apply the boolean modifier
            apply_bool_mod(context, bool_cells)
            # transfer the normals
            if group.transfer_normals:
                apply_data_transfer_mod(context, bool_cells, ob)
            # delete the invalid cells
            cells = collection.objects[:]
            shards = len(points)
            # cleanup the cells
            for i, cell in enumerate(cells):
                cell.select_set(True)
                if cell in bool_cells:
                    continue
                if points[i] in points_inside:
                    faces = cell.data.polygons
                    faces.foreach_set('hide', [False,]*len(faces))
                    if add_interior_v_group:
                        indexs = list(range(len(cell.data.vertices)))
                        cell.vertex_groups[0].add(indexs, 1.0, 'ADD')
                    continue
                try:
                    bpy.data.meshes.remove(cell.data)
                    bpy.data.objects.remove(cell)
                except:
                    pass
                shards -= 1
            # set the origin to geometry
            bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY', center='MEDIAN')
            # print info
            t = round(time.time() - s, 2)
            print('ALL DONE!', shards, 'Shards, in', t, 'Seconds')
            print(f'{"#"*74}\n')
        #restore random seed
        group.random_seed = random_seed
        return {'FINISHED'}

############################################################################################
####################################### The UI #############################################
############################################################################################
# Voronoi Fracturing Panel
class ANIMAX_PT_voronoi_fracture(Panel):
    bl_label = "Voronoi Fracture"
    bl_parent_id = "ANIMAX_PT_main"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_category = "ANIMAX"
    
    @classmethod
    def poll(self, context):
        if context.scene.animax_tabs != 'FRACTURE':
            return False
        return context.scene.animax_fracture_tabs == 'VORONOI'
        
    def draw(self, context):
        scn = context.scene
        group = scn.animax_voronoi_fracture
        layout = self.layout
        # Points Panel
        box = layout.box()
        box.prop(group, "expand_panel", index = 0,
        icon="TRIA_DOWN" if group.expand_panel[0] else "TRIA_RIGHT",
        icon_only=True, emboss = False, text = 'Points')
        if group.expand_panel[0]:
            row = box.row()
            row.prop(group, 'method', expand = True)
            box.prop(group, 'points_count')
            if group.method == 'QUICK':
                box.prop(group, 'distribution', text = '')
                if group.distribution == 'GRID':
                    box.prop(group, 'grid_random', slider = True)
                box.prop(group, 'random_seed')
            else:
                row = box.row()
                row.prop(group, 'points_source')
                if 'VERTS' in group.points_source:
                    box.prop(group, 'collection', text = '')
            box.prop(group, 'use_annotations')
            if group.use_annotations:
                box.prop(group, 'annotation_percentage')
        # Shards Panel        
        box = layout.box()
        box.prop(group, "expand_panel", index = 1,
        icon="TRIA_DOWN" if group.expand_panel[1] else "TRIA_RIGHT",
        icon_only=True, emboss = False, text = 'Shards')
        if group.expand_panel[1]:
            box.prop(group, 'margin')
            box.prop(group, 'material_index')
            row = box.row()
            row.prop(group, 'scale')
            col = box.column()
            box2 = col.box()
            box2.prop(group, "expand_panel", index = 2,
            icon="TRIA_DOWN" if group.expand_panel[2] else "TRIA_RIGHT",
            icon_only=True, emboss = False, text = 'Boolean')
            if group.expand_panel[2]:
                row = box2.row()
                row.prop(group, 'bool_solver', expand = True)
                col2 = box2.column(align=True)
                if group.bool_solver == 'EXACT':
                    col2.prop(group, 'use_hole_tolerant')
                    col2.prop(group, 'use_self')
                else:
                    col2.prop(group, 'overlap_threshold')
            box3 = col.box()
            box3.prop(group, "expand_panel", index = 3,
            icon="TRIA_DOWN" if group.expand_panel[3] else "TRIA_RIGHT",
            icon_only=True, emboss = False, text = 'Extra Options')
            if group.expand_panel[3]:
                col3 = box3.column(align=True)
                col3.prop(group, 'interior_v_group')
                col3.prop(group, 'smooth_interior')
                col3.prop(group, 'transfer_normals')
                col3.prop(group, 'sharp_edges')
                col3.prop(group, 'split_sharp')
                
        len_obs = len(context.selected_editable_objects)
        if len_obs > 1:
            col = layout.column()
            col.alert = True
            col.label(text = 'Multiple Objects Selected', icon = 'INFO')
            col.prop(group, 'randomize_seed')
        col = layout.column()
        col.alert = len_obs > 1
        col.scale_y = 1.75
        col.operator("animax.voronoi_fracture")
    
############################################################################################
################################ Register/Unregister #######################################
############################################################################################
    
classes = (    
    VoronoiFractureGroup,
    ANIMAX_OT_voronoi_fracture,
    ANIMAX_PT_voronoi_fracture,
)

def register():
    from bpy.utils import register_class
    for cls in classes:
        register_class(cls)
        
    bpy.types.Scene.animax_voronoi_fracture = PointerProperty(type = VoronoiFractureGroup)

def unregister():
    from bpy.utils import unregister_class
    for cls in reversed(classes):
        unregister_class(cls)
        
    del bpy.types.Scene.animax_voronoi_fracture

if __name__ == "__main__":
    register()